#!/usr/bin/python3.7
########################################################################################
# weather.py - Computs daily weather characteristics.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The weather module for this PV-T model.

This module computes daily solar irradiance, based on the time of year, lattitude etc.,
and factors in the time of day, and cloud cover, to give an accurate estimate of the
solar irradiance at any given time.

Extensibly, it has the potential to compute the rainfail in mm/time_step s.t. the
cooling effect on panels can be estimated and included into the model as well.

"""

import calendar
import collections
import datetime
import logging
import math
import os
import random

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Set, Tuple, Union

import json
import pysolar

from ..__utils__ import MissingParametersError, read_yaml


from .constants import (
    ZERO_CELCIUS_OFFSET,
)
from .__utils__ import (
    BaseDailyProfile,
    Date,
    ProgrammerJudgementFault,
    PVT_SYSTEM_MODEL_LOGGER_NAME,
    WeatherConditions,
)

__all__ = ("WeatherForecaster",)


# A parameter used in modelling weather data curves.
G_CONST = 1

# The resolution to which random numbers are generated
RAND_RESOLUTION = 100

logger = logging.getLogger(PVT_SYSTEM_MODEL_LOGGER_NAME)


class _DailyProfile(BaseDailyProfile):
    """
    Represents a day's solar irradiance profile.

    """

    @property
    def profile(self) -> Dict[datetime.time, float]:
        """
        Publically yields the internal "_profile" for average purposes.

        :return:
            The profile as a mapping between time and value.

        """

        return self._profile

    def __add__(self, other) -> Any:  # type: ignore
        """
        Adds two profiles together and returns the result.

        :param other:
            The other instance of a :class:`_DailySolarIrradianceProilfe` to which this
            one should be add, then the two returned.

        :return:
            A new :class:`_DailyProfile` instance, instantiated with the
            profile made by combining the two profiles.

        """

        profile = {
            key: self_value + other.profile[key]
            for key, self_value in self.profile.items()
        }

        return _DailyProfile(profile)

    def __truediv__(self, divisor: float) -> Any:  # type: ignore
        """
        Divides every value in the profile by the divisor.

        :param divisor:
            The number by which every value in the profile should be divided.

        :return:
            A new :class:`_DailyProfile` instance, instantiated with the
            profile made by dividing the current profile by the divisor.

        """

        profile = {key: value / divisor for key, value in self.profile.items()}

        return _DailyProfile(profile)


@dataclass
class _MonthlyWeatherData:
    """
    Contains weather data for a month.

    .. attribute:: num_days
        The average number of days in the month.

    .. attribute:: cloud_cover
        The probabilty that any given day within the month is cloudy.

    .. attribute:: rainy_days
        The average number of days in the month for which rain occurs.

    .. attribute:: day_temp
        The average daytime temperature, measured in Kelvin, for the month.

    .. attribute:: night_temp
        The average nighttime temperature, measured in Kelvin, for the month.

    .. attribute:: solar_irradiance_profiles
        A mapping between the day in the month and the solar irradiance profile for
        that day.

    .. attribute:: override_irradiance_profile
        If set, this will override the default behaviour of searching for the internal
        profile for the given day, or computing the average irradiance profile, and will
        simply return the profile set on this attribute when called.

    .. attribute:: average_temperature_profile
        The average temperature profile for the month.

    .. attribute:: sunrise
        The sunrise time. Can be set later.

    .. attribute:: sunset
        The sunset time. Can be set later.

    """

    # Private Attributes:
    #
    # .. attribute:: _average_irradiance_profile
    #   Used to store the average solar irradiance profile for the month.
    #

    month_name: str
    num_days: float
    cloud_cover: float
    rainy_days: float
    day_temp: float
    night_temp: float
    solar_irradiance_profiles: Dict[int, _DailyProfile] = field(default_factory=dict)
    _average_irradiance_profile: _DailyProfile = _DailyProfile(dict())
    override_irradiance_profile: _DailyProfile = _DailyProfile(dict())
    average_temperature_profile: _DailyProfile = _DailyProfile(dict())
    sunrise: Optional[datetime.time] = None
    sunset: Optional[datetime.time] = None

    def __repr__(self) -> str:
        """
        Return a standard representation of the class.

        :return:
            A nicely-formatted monthly-weather data string.

        """

        return (
            "_MonthlyWeatherData(month: {}, num_days: {}, cloud_cover: {}, ".format(
                self.month_name,
                self.num_days,
                self.cloud_cover,
            )
            + "rainy_days: {}, day_temp: {}, night_temp: {}, ".format(
                self.rainy_days,
                self.day_temp,
                self.night_temp,
            )
            + "solar_profiles_set: [{}])".format(
                "X" if self.solar_irradiance_profiles is not None else " "
            )
        )

    @classmethod
    def from_yaml(
        cls, month_name: str, monthly_weather_data: Dict[str, Union[str, float]]
    ) -> Any:  # type: ignore
        """
        Checks for present fields and instantiates from YAML data.

        :param month_name:
            A `str` giving a three-letter representation of the month.

        :param monthly_weather_data:
            A `dict` containing the weather data for the month, extracted raw from the
            weather YAML file.

        :return:
            An instance of the class.

        """

        try:
            return cls(
                month_name,
                float(monthly_weather_data["num_days"]),
                float(monthly_weather_data["cloud_cover"]),
                float(monthly_weather_data["rainy_days"]),
                float(monthly_weather_data["day_temp"]) + ZERO_CELCIUS_OFFSET,
                float(monthly_weather_data["night_temp"]) + ZERO_CELCIUS_OFFSET,
            )
        except KeyError as e:
            raise MissingParametersError(
                "WeatherForecaster",
                "Missing fields in YAML file. Error: {}".format(str(e)),
            ) from None

    @property
    def average_irradiance_profile(self) -> _DailyProfile:
        """
        Returns the average daily solar irradiance profile for the month.

        :return:
            The average solar irradiance profile for the month as a
            :class:`_DailyProfile` instance.

        """

        if self.override_irradiance_profile is not None:
            return self.override_irradiance_profile

        if self._average_irradiance_profile is not None:
            return self._average_irradiance_profile

        # Instantiate a counter.
        counter: Dict[datetime.time, float] = collections.defaultdict(float)

        if self.solar_irradiance_profiles is None:
            raise ProgrammerJudgementFault(
                "The irradiance profile in the weather module should not be None."
            )

        # Loop through all internally-held profiles, and sum up the values at each given
        # time, dividing by the number of days in the month.
        for profile in self.solar_irradiance_profiles.values():
            for time, value in profile.profile.items():
                counter[time] += value / self.num_days

        # Store this as a profile on the class for future use, and return this profile.
        self._average_irradiance_profile = _DailyProfile(counter)

        return self._average_irradiance_profile


def _get_solar_angles(
    latitude: float, longitude: float, date_and_time: datetime.datetime
) -> Tuple[float, float]:
    """
    Determine the azimuthal_angle (right-angle) and declination of the sun.

    :param latitude:
        The latitude of the PV-T set-up.

    :param longitude:
        The longitude of the PV-T set-up.

    :param date_and_time:
        The current date and time.

    :return:
        A `tuple` containing the azimuthal angle and declination of the sun at the
        given date and time.

    """

    return (
        pysolar.solar.get_azimuth(latitude, longitude, date_and_time),
        pysolar.solar.get_altitude(latitude, longitude, date_and_time),
    )


def _get_sunrise(
    latitude: float,
    longitude: float,
    date_and_time: datetime.datetime,
) -> datetime.time:
    """
    Determine the sunrise time for the month.

    :param latitude:
        The latitude of the PV-T set-up.

    :param longitude:
        The longitude of the PV-T set-up.

    :param date_and_time:
        The current date and time.

    :return:
        The time of sunrise for the month, returned as a :class:`datetime.time`.

    """

    _, declination = _get_solar_angles(latitude, longitude, date_and_time)
    if declination > 0:
        return date_and_time.time()
    return _get_sunrise(
        latitude,
        longitude,
        date_and_time.replace(hour=date_and_time.hour + 1),
    )


def _get_sunset(
    latitude: float,
    longitude: float,
    date_and_time: datetime.datetime,
    declination=0,
) -> datetime.time:
    """
    Determine the sunset time for the month.

    :param latitude:
        The latitude of the PV-T set-up.

    :param longitude:
        The longitude of the PV-T set-up.

    :param date_and_time:
        The current date and time.

    :return:
        The time of sunset for the month, returned as a :class:`datetime.time`.

    """

    _, declination = _get_solar_angles(latitude, longitude, date_and_time)
    if declination > 0:
        return date_and_time.time()
    return _get_sunset(
        latitude,
        longitude,
        date_and_time.replace(hour=date_and_time.hour - 1),
    )


class WeatherForecaster:
    """
    Represents a weather forecaster, determining weather conditions and irradiance.

    .. attribute:: mains_water_temperature
        The mains water temperature, measured in Kelvin.

    """

    # Private attributes:
    #
    # .. attribute:: _average_irradiance
    #   A `bool` giving whether to average the solar irradiance intensities (True), or
    #   use the data for each day as called (False).
    #
    # .. attribute:: _average_air_pressure
    #   The average air pressure measured in Pascals.
    #
    # .. attribute:: _month_abbr_to_num
    #   A mapping from month abbreviated name (eg, "jan", "feb" etc) to the number of
    #   the month in the year.
    #
    # .. attribute:: _monthly_weather_data
    #   A `dict` mapping month number to :class:`_MonthlyWeatherData` instances
    #   containing weather information for that month.
    #
    # .. attribute:: _override_ambient_temperature
    #   The ambient temperature, measured in Kelvin, used as a constant value rather
    #   than time-s[ecific data.
    #
    # .. attribute:: _override_irradiance
    #   The solar irradiance, measured in Watts per meter squared, to use as a constant
    #   value rather than time-specific data.
    #
    # .. attribute:: _override_wind_speed
    #   The wind speed, measured in meters per second, used as a constant value rather
    #   than time-specific data.
    #
    # .. attribute:: _solar_insolation
    #   The solar insolation, measured in Watts per meter squared, that would hit the
    #   UK on a clear day with no other factors present.
    #

    _month_abbr_to_num = {
        name.lower(): num
        for num, name in enumerate(calendar.month_abbr)
        if num is not None
    }

    def __init__(
        self,
        ambient_tank_temperature: float,
        average_air_pressure: float,
        average_irradiance: bool,
        mains_water_temperature: float,
        monthly_weather_data: Dict[str, Dict[str, Union[str, float]]],
        monthly_irradiance_profiles: Dict[Date, _DailyProfile],
        monthly_temperature_profiles: Dict[Date, _DailyProfile],
        override_ambient_temperature: Optional[float],
        override_irradiance: Optional[float],
        overridw_wind_speed: Optional[float],
    ) -> None:
        """
        Instantiate a weather forecaster class.

        :param ambient_tank_temperature:
            The average ambient temperature surrounding the hot-water wank.

        :param average_air_pressure:
            The average air pressure, measured in Pascals.

        :param average_irradiance:
            Whether to average the solar intensity for each month (True), or use the
            data for each day (False) as required.

        :param mains_water_temperature:
            The mains water temperature, measured in Kelvin.

        :param monthly_weather_data:
            The monthly weather data, extracted raw from the weather data YAML file.

        :param monthly_irradiance_profiles:
            A mapping between :class:`__utils__.Date` instances and solar irradiance
            profiles stored as :class:`_DailyProfile` instances.

        :param override_ambient_temperature:
            Used to override the ambient temperature profiles with a constant value.

        :param override_irradiance:
            Used to override the solar-irradiance profiles with a constant value.

        :param override_wind_speed:
            Used to override the wind-speed profiles with a constant value.

        """

        self._average_air_pressure = average_air_pressure
        self._average_irradiance = average_irradiance
        self._override_ambient_temperature = override_ambient_temperature
        self._override_irradiance = override_irradiance
        self._override_wind_speed = overridw_wind_speed
        self.ambient_tank_temperature = ambient_tank_temperature
        self.mains_water_temperature = mains_water_temperature + ZERO_CELCIUS_OFFSET

        self._monthly_weather_data: Dict[int, _MonthlyWeatherData] = {
            self._month_abbr_to_num[month]: _MonthlyWeatherData.from_yaml(
                month, month_data  # type: ignore
            )
            for month, month_data in monthly_weather_data.items()
        }

        for month, weather_data in self._monthly_weather_data.items():
            weather_data.solar_irradiance_profiles = {
                date.day: profile
                for date, profile in monthly_irradiance_profiles.items()
                if date.month == month
            }
            if len(weather_data.solar_irradiance_profiles) == 1:
                weather_data.override_irradiance_profile = next(
                    iter(weather_data.solar_irradiance_profiles.values())
                )
            try:
                weather_data.average_temperature_profile = monthly_temperature_profiles[
                    Date(1, month)
                ]
            except KeyError as e:
                logger.debug(
                    "Unable to set average temperature profile for month %s; "
                    "data not present: %s",
                    month,
                    str(e),
                )
                continue

    def __repr__(self) -> str:
        """
        Return a nice-looking representation of the weather forecaster.

        :return:
            A nicely-formatted string giving information about the weather forecaster.

        """

        return (
            "WeatherForecaster("
            f"_average_irradiance: {self._average_irradiance}, "
            f"mains_water_temperature: {self.mains_water_temperature}K, "
            f"num_months: {len(self._monthly_weather_data.keys())}, "
            f"override_ambient_temperature: {self._override_ambient_temperature}K, "
            f"override_irradiance: {self._override_irradiance}W/m^2, "
            f"overridw_wind_speed: {self._override_wind_speed}m/s"
            ")"
        )

    def _ambient_temperature(
        self, latitude: float, longitude: float, date_and_time: datetime.datetime
    ) -> float:
        """
        Return the ambient temperature, in Kelvin, based on the date and time.

        A sine curve is fitted, and the temp extracted.

        The models used in this function are obtained, with permission, from
        https://mathscinotes.com/wp-content/uploads/2012/12/dailytempvariation.pdf
        and use an improved theoretical model from a previous paper.

        :param latitude:
            The latitude of the set-up.

        :param longitude:
            The longitude of the set-up.

        :param date_and_time:
            The current date and time.

        :return:
            The temperature in Kelvin.

        """

        max_temp = self._monthly_weather_data[date_and_time.month].day_temp
        min_temp = self._monthly_weather_data[date_and_time.month].night_temp
        temp_range = max_temp - min_temp

        if (
            self._monthly_weather_data[date_and_time.month].sunrise is None
            or self._monthly_weather_data[date_and_time.month].sunset is None
        ):
            self._monthly_weather_data[date_and_time.month].sunrise = _get_sunrise(
                latitude, longitude, date_and_time.replace(hour=0)
            )
            self._monthly_weather_data[date_and_time.month].sunset = _get_sunset(
                latitude, longitude, date_and_time.replace(hour=23)
            )

        return (
            temp_range
            * math.exp(-(date_and_time.hour + date_and_time.minute / 60 - 12) * G_CONST)
            * (1 + (date_and_time.hour + date_and_time.minute / 60 - 12) / 12)
            ** (G_CONST * 12)
            + min_temp
        )

    def _cloud_cover(
        self, cloud_efficacy_factor: float, date_and_time: datetime.datetime
    ) -> float:
        """
        Computes the cloud clover based on the time of day and various factors.

        :param cloud_efficacy_factor:
            The extend to which cloud cover affects the sunlight. This is multiplied by
            a random number which further reduces the effect of the cloud cover in
            reducing the sunlight.

        :param date_and_time:
            The date and time of day, used to determine which month the cloud cover
            should be retrieved for.

        :return:
            The fractional effect that the cloud cover has to reduce

        """

        # Extract the cloud cover probability for the month.
        cloud_cover_prob = self._monthly_weather_data[date_and_time.month].cloud_cover

        # Generate a random number between 1 and 0 for that month based on this factor.
        rand_prob: float = random.randrange(0, RAND_RESOLUTION, 1) / RAND_RESOLUTION

        # Determine what effect the cloudy (or not cloudy) conditions have on the solar
        # insolation. Generate a fractional reduction based on this.
        # Return this number
        return cloud_cover_prob * rand_prob * cloud_efficacy_factor

    def _irradiance(self, date_and_time: datetime.datetime) -> float:
        """
        Return the solar irradiance, in Watts per meter squared, at the current date and
        time.

        :param date_and_time:
            The date and time.

        :return:
            The solar irradiance, in Watts per meter squared.

        """

        if self._average_irradiance:
            return self._monthly_weather_data[
                date_and_time.month
            ].average_irradiance_profile[date_and_time.time()]

        return self._monthly_weather_data[
            date_and_time.month
        ].solar_irradiance_profiles[date_and_time.day][date_and_time.time()]

    @classmethod
    def from_data(  # pylint: disable=too-many-branches
        cls,
        average_irradiance: bool,
        solar_irradiance_filenames: Set[str],
        temperature_filenames: Set[str],
        weather_data_path: str,
        override_ambient_temperature: Optional[float] = None,
        override_irradiance: Optional[float] = None,
        override_wind_speed: Optional[float] = None,
        use_pvgis: bool = False,
    ) -> Any:  # type: ignore
        """
        Instantiate a :class:`WeatherForecaster` from paths to various data files.

        :param average_irradiance:
            Whether to average the solar irradiances wihtin each month (True), or use
            the value for each day specifically when called (False).

        :param solar_irradiance_filenames:
            A `set` of paths to files containing solar irradiance profiles. The format
            of these files is in the form as given by the PVGIS (Photovoltaic,
            Geographic Information Service). Processing of these files occurs here.

        :param temperature_filenames:
            A `set` of paths to files containing temperature profiles.

        :param weather_data_path:
            The path to the weather-data file. This contains basic information about the
            location that is weather-related.

        :param override_ambient_temperature:
            If specified, the value will be used to override the internal ambient
            temperature profile(s). The value should be in degrees Kelvin.

        :param override_irradiance:
            If specified, the value will be used to override the internal solar
            irradiance profile(s).

        :param override_wind_speed:
            If specified, the value will be used to override the internal wind-speed
            profile(s).

        :param use_pvgis:
            Whether to use data obtained from PVGIS (True) or not (False).

        :return:
            A :class:`WeatherForecaster` instance.

        """

        data = read_yaml(weather_data_path)

        try:
            data.pop("solar_insolation")
        except KeyError:
            logger.error(
                "Weather forecaster from %s is missing 'solar_insolation' data.",
                weather_data_path,
            )
            raise MissingParametersError(
                "WeatherForecaster",
                "The solar insolation param is missing from {}.".format(
                    weather_data_path
                ),
            ) from None

        try:
            mains_water_temp = data.pop("mains_water_temp")
        except KeyError:
            logger.error(
                "Weather forecaster from %s is missing 'mains_water_temp' data.",
                weather_data_path,
            )
            raise MissingParametersError(
                "WeatherForecaster",
                "The mains water temperature param is missing from {}.".format(
                    weather_data_path
                ),
            ) from None

        try:
            ambient_tank_temperature = (
                data.pop("average_ambient_household_temperature") + ZERO_CELCIUS_OFFSET
            )
        except KeyError:
            logger.error(
                "Weather forecaster from %s is missing "
                "'average_ambient_household_temperature' data.",
                weather_data_path,
            )
            raise MissingParametersError(
                "WeatherForecaster",
                "The average ambient tank temperature param is missing from {}.".format(
                    weather_data_path
                ),
            ) from None

        try:
            average_air_pressure = data.pop("average_air_pressure")
        except KeyError:
            logger.error(
                "Weather forecaster from %s is missing " "'average_air_pressure' data.",
                weather_data_path,
            )
            raise MissingParametersError(
                "WeatherForecaster",
                "The average air pressure param is missing from {}.".format(
                    weather_data_path
                ),
            ) from None

        # NOTE:
        # > Here, PVGIS code exists in a block s.t. the override profiles obtained from
        # > Maria's paper using a highly scientific method :p can be used instead.

        # Loop through all data files, reading in the data, and adding to a profile
        # keyed only by month and day.
        if use_pvgis:
            temp_irradiance_profiles: Dict[Date, dict] = collections.defaultdict(dict)

            for filename in solar_irradiance_filenames:
                with open(filename) as f:
                    filedata = json.load(f)
                processed_filedata: Dict[datetime.datetime, float] = {
                    datetime.datetime.strptime(entry["time"], "%Y%m%d:%H%M"): entry[
                        "G(i)"
                    ]
                    for entry in filedata["outputs"]["hourly"]
                }

                num_years = len({date.year for date in processed_filedata})

                for key, value in processed_filedata.items():
                    # This is being wrapped in a try-except block to allow for assigning of
                    # both items that are already present, and those which aren't.
                    try:
                        temp_irradiance_profiles[Date.from_date(key.date())][
                            key.time()
                        ] += (value / num_years)
                    except KeyError:
                        temp_irradiance_profiles[Date.from_date(key.date())][
                            key.time()
                        ] = (value / num_years)

            # These profiles now need to be cast to _DailyProfiles
            monthly_irradiance_profiles: Dict[Date, _DailyProfile] = {
                date: _DailyProfile(profile)
                for date, profile in temp_irradiance_profiles.items()
            }
        else:
            # Cycle through the various profile files, opening the profiles and storing as a
            # mapping.
            monthly_irradiance_profiles = dict()
            for filename in solar_irradiance_filenames:
                with open(filename, "r") as f:
                    filedata = json.load(f)

                # Process the profile and store it.
                try:
                    monthly_irradiance_profiles[
                        Date(1, cls._month_abbr_to_num[os.path.basename(filename)[:3]])
                    ] = _DailyProfile(
                        {
                            datetime.datetime.strptime(key, "%H:%M").time(): value
                            for key, value in filedata.items()
                        }
                    )
                except ValueError:
                    logger.info(
                        "Weather data contains additional information. Attempting to use convert."
                    )
                    monthly_irradiance_profiles[
                        Date(1, cls._month_abbr_to_num[os.path.basename(filename)[:3]])
                    ] = _DailyProfile(
                        {
                            datetime.datetime.strptime(key, "%H:%M:%S").time(): value
                            for key, value in filedata.items()
                        }
                    )

        temperature_profiles: Dict[Date, _DailyProfile] = dict()
        for filename in temperature_filenames:
            with open(filename, "r") as f:
                filedata = json.load(f)

            # Process the profile and store it.
            try:
                temperature_profiles[
                    Date(1, cls._month_abbr_to_num[os.path.basename(filename)[:3]])
                ] = _DailyProfile(
                    {
                        datetime.datetime.strptime(key, "%H:%M").time(): value
                        + ZERO_CELCIUS_OFFSET
                        for key, value in filedata.items()
                    }
                )
            except ValueError:
                logger.info(
                    "Weather data contains additional information. Attempting to use convert."
                )
                temperature_profiles[
                    Date(1, cls._month_abbr_to_num[os.path.basename(filename)[:3]])
                ] = _DailyProfile(
                    {
                        datetime.datetime.strptime(key, "%H:%M:%S").time(): value
                        + ZERO_CELCIUS_OFFSET
                        for key, value in filedata.items()
                    }
                )

        # # @@@ For now, the solar irradiance profiles and temperature profiles for the
        # # @@@ missing months are filled in here.
        # monthly_irradiance_profiles[Date(1, 2)] = monthly_irradiance_profiles[
        #     Date(1, 1)
        # ]
        # monthly_irradiance_profiles[Date(1, 3)] = monthly_irradiance_profiles[
        #     Date(1, 4)
        # ]
        # monthly_irradiance_profiles[Date(1, 5)] = monthly_irradiance_profiles[
        #     Date(1, 4)
        # ]
        # monthly_irradiance_profiles[Date(1, 6)] = monthly_irradiance_profiles[
        #     Date(1, 8)
        # ]
        # monthly_irradiance_profiles[Date(1, 7)] = _DailyProfile(
        #     {
        #         key: (
        #             monthly_irradiance_profiles[Date(1, 8)].profile[key] * 3
        #             + monthly_irradiance_profiles[Date(1, 4)].profile[key]
        #         )
        #         / 4
        #         for key in monthly_irradiance_profiles[Date(1, 8)].profile.keys()
        #     }
        # )
        # monthly_irradiance_profiles[Date(1, 9)] = monthly_irradiance_profiles[
        #     Date(1, 4)
        # ]
        # monthly_irradiance_profiles[Date(1, 10)] = monthly_irradiance_profiles[
        #     Date(1, 4)
        # ]
        # monthly_irradiance_profiles[Date(1, 11)] = monthly_irradiance_profiles[
        #     Date(1, 1)
        # ]
        # monthly_irradiance_profiles[Date(1, 12)] = monthly_irradiance_profiles[
        #     Date(1, 1)
        # ]

        # temperature_profiles[Date(1, 2)] = temperature_profiles[Date(1, 1)]
        # temperature_profiles[Date(1, 3)] = temperature_profiles[Date(1, 4)]
        # temperature_profiles[Date(1, 5)] = temperature_profiles[Date(1, 4)]
        # temperature_profiles[Date(1, 6)] = temperature_profiles[Date(1, 7)]
        # temperature_profiles[Date(1, 9)] = temperature_profiles[Date(1, 4)]
        # temperature_profiles[Date(1, 10)] = temperature_profiles[Date(1, 4)]
        # temperature_profiles[Date(1, 11)] = temperature_profiles[Date(1, 1)]
        # temperature_profiles[Date(1, 12)] = temperature_profiles[Date(1, 1)]

        # Instantiate and return a Weather Forecaster based off of this weather data.
        return cls(
            ambient_tank_temperature,
            average_air_pressure,
            average_irradiance,
            mains_water_temp,
            data,
            monthly_irradiance_profiles,
            temperature_profiles,
            override_ambient_temperature + ZERO_CELCIUS_OFFSET
            if override_ambient_temperature is not None
            else None,
            override_irradiance,
            override_wind_speed,
        )

    def get_weather(
        self,
        latitude: float,
        longitude: float,
        cloud_efficacy_factor: float,  # pylint: disable=unused-argument
        date_and_time: Optional[datetime.datetime] = None,
    ) -> WeatherConditions:
        """
        Computes the solar irradiance based on weather conditions at the time of day.

        :param latitude:
            The latitude of the PV-T set-up.

        :param longitude:
            The longitude of the PV-T set-up.

        :param cloud_efficacy_factor:
            The extend to which cloud cover affects the sunlight. This is multiplied by
            a random number which further reduces the effect of the cloud cover in
            reducing the sunlight.

        :param date_and_time:
            The date and time of day, used to calculate the irradiance.

        :return:
            A :class:`__utils__.WeatherConditions` giving the solar irradiance, in
            watts per meter squared, and the angles, both azimuthal and declination, of
            the sun's position in the sky.

        """

        # Factor in the weather conditions and cloud cover to compute the current solar
        # irradiance.
        if date_and_time is not None:
            # Based on the time, compute the sun's position in the sky, making sure to
            # account for the seasonal variation.
            azimuthal_angle, declination = _get_solar_angles(
                latitude, longitude, date_and_time
            )
            irradiance: float = self._irradiance(date_and_time)  # * (
            # 1 - self._cloud_cover(cloud_efficacy_factor, date_and_time)
            # )
            # >>> The ambient temperature is now determined from a temperature profile.
            # # Compute the ambient temperature.
            # ambient_temperature: float = self._ambient_temperature(
            #     latitude, longitude, date_and_time
            # )
            # <<< >>> Ambient temperature determination using temperature profile.
            ambient_temperature: float = self._monthly_weather_data[
                date_and_time.month
            ].average_temperature_profile[date_and_time.time()]
            # <<< E.O. code-alteration block
        else:
            azimuthal_angle = 180
            declination = 45
            if self._override_irradiance is None:
                raise MissingParametersError(
                    "WeatherForecaster",
                    "The weather forecaster was called for time-independant weather "
                    "data without an override irradiance provided.",
                )
            irradiance = self._override_irradiance
            if self._override_ambient_temperature is None:
                raise MissingParametersError(
                    "WeatherForecaster",
                    "The weather forecaster was called for time-independant weather "
                    "data without an override ambient temperature provided.",
                )
            ambient_temperature = self._override_ambient_temperature

        # * Compute the wind speed
        if self._override_wind_speed is not None:
            wind_speed: float = self._override_wind_speed  # [m/s]
        else:
            wind_speed = 5

        # Return all of these in a WeatherConditions variable.
        return WeatherConditions(
            _irradiance=irradiance,
            azimuthal_angle=azimuthal_angle,
            ambient_tank_temperature=self.ambient_tank_temperature,
            ambient_temperature=ambient_temperature,
            declination=declination,
            mains_water_temperature=self.mains_water_temperature,
            pressure=self._average_air_pressure,
            wind_speed=wind_speed,
        )



#!/usr/bin/python3.7
# type: ignore
########################################################################################
# validation_analysis.py - The glazing analysis component for the model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################
"""
Used for analysis of various differences between outputs from model runs.

NOTE: The mypy type checker is instructed to ignore this component. This is done due to
the lower standards applied to the analysis code, and the failure of mypy to correctly
type-check the external matplotlib.pyplot module.

"""

import argparse
import collections
import os
import sys

from logging import Logger
from typing import Any, List, Dict, Optional, Set

import re

import numpy
import yaml

from matplotlib import pyplot as plt

try:
    from ..__utils__ import get_logger
    from ..pvt_system.constants import (  # pylint: disable=unused-import
        HEAT_CAPACITY_OF_WATER,
    )
    from .__utils__ import (
        GraphDetail,
        load_model_data,
        plot_figure,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise

__all__ = ("analyse",)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"
# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"
# Used to distinguish dynamic data sets.
DYNAMIC_DATA_TYPE = "dynamic"
# The directory into which which should be saved
NEW_FIGURES_DIRECTORY: str = "figures"
# Used to identify the "time" data set.
TIME_KEY = "time"
# The directory in which old figures are saved and stored for long-term access
OLD_FIGURES_DIRECTORY: str = "old_figures"
# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"
# Name of the directory containing validation data.
VALIDATION_DATA_DIRECTORY = "validation_data"
# How detailed the graph should be
GRAPH_DETAIL: GraphDetail = GraphDetail.lowest
# How many values there should be between each tick on the x-axis
# X_TICK_SEPARATION: int = int(8 * GRAPH_DETAIL.value / 48)
X_TICK_SEPARATION: int = 8
# Which days of data to include
DAYS_TO_INCLUDE: List[bool] = [False, True]


def _parse_args(args) -> argparse.Namespace:
    """
    Parse the CLI args.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data-file-directory",
        "-dfdir",
        help="Path to the directory containing data files to parse.",
    )
    parser.add_argument(
        "--show-output",
        "-so",
        action="store_true",
        default=False,
        help="Show the output figures generated.",
    )

    return parser.parse_args(args)


def _resolution_from_graph_detail(
    graph_detail: GraphDetail, num_data_points: int
) -> int:
    """
    Determine the x-step resolution of the plot.

    I realise here that the word "resolution" is over-used. Here, the number of data
    points (in the system data) that need to be absorbed into one graph point is
    calculated and returned.

    :param graph_detail:
        The detail level needed on the graph.

    :param num_data_points:
        The number of data points recorded in the plot.

    :return:
        The number of data points that need to be absorbed per graph point.

    """

    # * For "lowest", include one point every half-hour.
    if graph_detail == GraphDetail.highest:
        return 1

    return int(num_data_points / graph_detail.value)


def _post_process_data(
    data_to_post_process: Dict[str, Dict[Any, Any]]
) -> Dict[str, Dict[Any, Any]]:
    """
    Carries out post-processing on data where necessary.

    Things to be computed:
        - Bulk Water Temperautre
        - Litres consumed

    :param data_to_post_process:
        The data to be post processed.

    :return:
        The post-processed data.

    """

    # * Cycle through all the data points and compute the new values as needed.
    for data_entry in data_to_post_process.values():
        data_entry["absorber_temperature_gain"] = (
            data_entry["collector_output_temperature"]
            - data_entry["collector_input_temperature"]
        )
    return data_to_post_process


def _validation_figure(
    data: Dict[Any, Dict[str, Any]],
    logger: Logger,
    prefix: str,
    regex: re.Pattern,
    variable_name: str,
    *,
    validation_filename: Optional[str] = None,
) -> None:
    """
    Carries out analysis on a validation figure.

    Based on the parameters passed in, a sub-set of files is selected for, Ilaria's
    data is parsed and plotted if relevant, and a figure is saved.

    :param data:
        A `dict` mapping filenames to raw JSON data extracted from the files.

    :param logger:
        The logger to use for the run.

    :param prefrix:
        A `str` to prepend to all filenames so that the output figures can be uniquely
        identified.

    :param regex:
        A compiled :class:`re.Pattern` to use for matching and variable parsing from
        filenames.
        NOTE: The regex should contain:
        - either: the groups "first_digit" and "second_digit" which, together, give the
          value of the variable that identifies the data set,
        - or: the group "variable_value" which gives the value of the variable that
          identifies the data set.

    :param variable_name:
        The name of the variable being experimented with. This is used for labelling
        data sets and trendlines.

    :param validation_filename:
        If specified, this gives the name of the validation data file from which
        validation data should be plotted. If `None`, then no validation data is parsed
        and plotted.

    """

    # Plot the glass-emissivity effect on a single-glazed Ilaria.
    reduced_data: Dict[str, Any] = collections.defaultdict(dict)
    thermal_efficiency_labels: Set[str] = set()
    electrical_efficiency_labels: Set[str] = set()
    for key, sub_dict in data.items():
        match = re.match(regex, key)
        if match is None:
            continue
        try:
            variable_value = float(
                f"{match.group('first_digit')}.{match.group('second_digit')}"
            )
        except IndexError:
            variable_value = match.group("variable_value")
        # Only plot significant portions covered.
        for sub_key, value in sub_dict.items():
            # Generate a unique indentifier string.
            file_identifier = f"{variable_name}={variable_value}"
            thermal_efficiency_identifier = f"{file_identifier} (therm.)"
            electrical_efficiency_identifier = f"{file_identifier} (elec.)"
            # Store the specific thermal efficiency.
            reduced_data[sub_key][thermal_efficiency_identifier] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(thermal_efficiency_identifier)
            # Store the specific electrical efficiency.
            reduced_data[sub_key][electrical_efficiency_identifier] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(electrical_efficiency_identifier)
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    # Thermal efficiency plot.
    logger.info("Plotting thermal efficiency against the reduced temperature.")
    _, ax1 = plt.subplots()

    # Plot Ilaria's data to validate against if relevant.
    if validation_filename is not None:
        # Parse the thermal-efficiency data.
        with open(
            os.path.join(VALIDATION_DATA_DIRECTORY, validation_filename),
            "r",
        ) as f:  #
            validation_data = yaml.safe_load(f)

        validation_data_dict = {
            entry["label"]: entry["thermal_efficiency"] for entry in validation_data
        }

        x_data_series = numpy.linspace(-0.005, 0.035, 100)
        y_data_sets = dict()
        for data_entry_name in ["single_glazed", "double_glazed", "unglazed"]:
            data_entry = validation_data_dict[data_entry_name]
            y_data_sets[data_entry_name] = (
                data_entry["x_squared"] * x_data_series ** 2
                + data_entry["linear_x"] * x_data_series
                + data_entry["y_intercept"]
            )

        for data_entry_name in sorted(["double_glazed", "single_glazed", "unglazed"]):
            # Plot the experimental data.
            ax1.plot(
                x_data_series,
                y_data_sets[data_entry_name],
            )

    ax1.set_prop_cycle(None)

    # Thermal efficiency plot.
    logger.info("Plotting thermal efficiency against the reduced temperature.")
    plot_figure(
        f"{prefix}_thermal_efficiency_against_reduced_temperature",
        reduced_data,
        first_axis_things_to_plot=thermal_efficiency_labels,
        first_axis_label="Thermal efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Thermal efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
        override_axis=ax1,
    )

    # Plot the electrical efficiency against the reduced temperature.
    logger.info("Plotting electrical efficiency against the reduced temperature.")

    plot_figure(
        f"{prefix}_electrical_efficiency_against_reduced_temperature",
        reduced_data,
        first_axis_things_to_plot=electrical_efficiency_labels,
        first_axis_label="Electrical efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Electrical efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
    )


def analyse_decoupled_steady_state_data(  # pylint: disable=too-many-branches
    data: Dict[Any, Dict[str, Any]], logger: Logger
) -> None:
    """
    Carry out analysis on a series of of steady-state data sets.

    :param data:
        The data to analyse. This should be a `dict` mapping the file name to the data
        extracted from the file.

    :param logger:
        The logger to use for the analysis run.

    """

    logger.info("Beginning steady-state analysis.")

    logger.info("Beginning autotherm glass-transmissivity analysis.")
    autotherm_glass_transmissivity_regex = re.compile(
        r"autotherm[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_glass_transmissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "autotherm_glass_transmissivity",
        autotherm_glass_transmissivity_regex,
        "glass_transmissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria glass-transmissivity analysis.")
    ilaria_glass_transmissivity_regex = re.compile(
        r"ilaria[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_glass_transmissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_glass_transmissivity",
        ilaria_glass_transmissivity_regex,
        "glass_transmissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning autotherm glass-emissivity analysis.")
    autotherm_glass_emissivity_regex = re.compile(
        r"autotherm[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_glass_emissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "autotherm_glass_emissivity",
        autotherm_glass_emissivity_regex,
        "glass_emissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria glass-emissivity analysis.")
    ilaria_glass_emissivity_regex = re.compile(
        r"ilaria[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_glass_emissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_glass_emissivity",
        ilaria_glass_emissivity_regex,
        "glass_emissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning autotherm pv-efficiency analysis.")
    autotherm_pv_efficiency_regex = re.compile(
        r"autotherm[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_pv_efficiency.*"
    )
    _validation_figure(
        data,
        logger,
        "autotherm_pv_efficiency",
        autotherm_pv_efficiency_regex,
        "pv_efficiency",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria pv-efficiency analysis.")
    ilaria_pv_efficiency_regex = re.compile(
        r"ilaria[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_pv_efficiency.*"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_pv_efficiency",
        ilaria_pv_efficiency_regex,
        "pv_efficiency",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning autotherm pv-emissivity analysis.")
    autotherm_pv_emissivity_regex = re.compile(
        r"autotherm[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_pv_emissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "autotherm_pv_emissivity",
        autotherm_pv_emissivity_regex,
        "pv_emissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria pv-emissivity analysis.")
    ilaria_pv_emissivity_regex = re.compile(
        r"ilaria[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_pv_emissivity.*"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_pv_emissivity",
        ilaria_pv_emissivity_regex,
        "pv_emissivity",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning autotherm thermal coefficient analysis.")
    autotherm_thermal_coefficient_regex = re.compile(
        r"autotherm[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_thermal_coefficient.*"
    )
    _validation_figure(
        data,
        logger,
        "autotherm_thermal_coefficient",
        autotherm_thermal_coefficient_regex,
        "thermal_coefficient",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria thermal coefficient analysis.")
    ilaria_thermal_coefficient_regex = re.compile(
        r"ilaria[^\d]*(?P<first_digit>[\d]*)_(?P<second_digit>[\d]*)_thermal_coefficient.*"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_thermal_coefficient",
        ilaria_thermal_coefficient_regex,
        "thermal_coefficient",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning autotherm glazing analysis.")
    autotherm_glazing_regex = re.compile(r"autotherm_(?P<variable_value>.*)_pc_1_0_.*")
    _validation_figure(
        data,
        logger,
        "autotherm_glazing",
        autotherm_glazing_regex,
        "glazing",
    )
    logger.info("Done.")
    plt.close()

    logger.info("Beginning Ilaria glazing analysis.")
    ilaria_glazing_regex = re.compile(r"ilaria_(?P<variable_value>.*)_pc_1_0_.*")
    _validation_figure(
        data,
        logger,
        "ilaria_glazing",
        ilaria_glazing_regex,
        "glazing",
        validation_filename="glazing_comparison.yaml",
    )
    logger.info("Done.")
    plt.close()

    logger.info("All validation analysis complete.")

    logger.info("Beginning Ilaria theoretical glazing analysis.")
    ilaria_glazing_regex = re.compile(
        r"ilarias_theoretical_panel_(?P<variable_value>.*).json"
    )
    _validation_figure(
        data,
        logger,
        "ilaria_glazing",
        ilaria_glazing_regex,
        "glazing",
        validation_filename="glazing_comparison.yaml",
    )
    logger.info("Done.")
    plt.close()

    logger.info("All validation analysis complete.")


def analyse(data_file_directory: str, show_output: Optional[bool] = False) -> None:
    """
    The main method for the analysis module.

    :param data_file_directory:
        The path to the directory containing data files to analyse.

    :param show_output:
        Whether to show the output files generated.

    """

    # * Set up the logger
    logger = get_logger(True, "pvt_analysis", True)

    # * Extract the data.
    data: Dict[str, Dict[str, Any]] = dict()
    for data_file_name in os.listdir(data_file_directory):
        data[data_file_name] = load_model_data(
            os.path.join(data_file_directory, data_file_name)
        )

        # If the data type is not decoupled, steady-state data, then exit.
        data_type = data[data_file_name].pop("data_type")
        if data_type not in (
            STEADY_STATE_DATA_TYPE,
            f"{DECOUPLED_DATA_TYPE}_{STEADY_STATE_DATA_TYPE}",
        ):
            logger.info(
                "Data type was neither 'dynamic' nor 'steady_state'. Omitting..."
            )
            data.pop(data_file_name)
            continue
        if "no_pv" in data_file_name:
            logger.info("Invalid data was removed.")
            data.pop(data_file_name)
            continue

    analyse_decoupled_steady_state_data(data, logger)

    logger.info("Analysis complete - all figures saved successfully.")

    if show_output:
        plt.show()


if __name__ == "__main__":
    parsed_args = _parse_args(sys.argv[1:])
    analyse(parsed_args.data_file_directory, parsed_args.show_output)



#!/usr/bin/python3.7
########################################################################################
# __main__.py - The main module for this, my first, PV-T model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The main module for the PV-T model.

This module coordinates the time-stepping of the itterative module, calling out to the
various modules where necessary, as well as reading in the various data files and
processing the command-line arguments that define the scope of the model run.

"""

import logging
import os

from typing import Dict, List, Optional, Set, Tuple

import numpy


from . import (
    exchanger,
    index_handler,
    load,
    mains_power,
    process_pvt_system_data,
    tank,
    weather,
)

from .runs import coupled, decoupled

from ..__utils__ import (  # pylint: disable=unused-import
    BColours,
    CarbonEmissions,
    get_logger,
    OperatingMode,
    SystemData,
    TemperatureName,
    TotalPowerData,
)
from .__utils__ import (
    ProgrammerJudgementFault,
    PVT_SYSTEM_MODEL_LOGGER_NAME,
    SOLAR_IRRADIANCE_FOLDERNAME,
    TEMPERATURE_FOLDERNAME,
    WEATHER_DATA_FILENAME,
)

from .constants import (
    DEFAULT_INITIAL_DATE_AND_TIME,
    ZERO_CELCIUS_OFFSET,
)


def _get_load_system(location: str) -> load.LoadSystem:
    """
    Instantiates a :class:`load.LoadSystem` instance based on the file data.

    :param locataion:
        The location being considered for which to instantiate the load system.

    :return:
        An instantiated :class:`load.LoadSystem` based on the input location.

    """

    load_profiles = {
        os.path.join(location, "load_profiles", filename)
        for filename in os.listdir(os.path.join(location, "load_profiles"))
    }
    load_system: load.LoadSystem = load.LoadSystem.from_data(load_profiles)
    return load_system


def _get_weather_forecaster(
    average_irradiance: bool,
    location: str,
    use_pvgis: bool,
    override_ambient_temperature: Optional[float],
    override_irradiance: Optional[float],
    override_wind_speed: Optional[float],
) -> weather.WeatherForecaster:
    """
    Instantiates a :class:`weather.WeatherForecaster` instance based on the file data.

    :param average_irradiance:
        Whether to use an average irradiance profile for the month (True) or use
        irradiance profiles for each day individually (False).

    :param location:
        The location currently being considered, and for which to instantiate the
        :class:`weather.WeatherForecaster` instance.

    :param override_ambient_temperature:
        Overrides the ambient temperature value. The value should be in degrees Celcius.

    :param override_irradiance:
        Overrides the irradiance value. The value should be in Watts per meter squared.

    :param override_wind_speed:
        Overrides the wind speed value. The value should be in meters per second.

    :param use_pvgis:
        Whether data from the PVGIS (Photovoltaic Geographic Information Survey) should
        be used (True) or not (False).

    :return:
        An instantiated :class:`weather.WeatherForecaster` instance based on the
        supplied parameters.

    """

    solar_irradiance_filenames = {
        os.path.join(location, SOLAR_IRRADIANCE_FOLDERNAME, filename)
        for filename in os.listdir(os.path.join(location, SOLAR_IRRADIANCE_FOLDERNAME))
    }
    temperature_filenames = {
        os.path.join(location, TEMPERATURE_FOLDERNAME, filename)
        for filename in os.listdir(os.path.join(location, TEMPERATURE_FOLDERNAME))
    }

    weather_forecaster: weather.WeatherForecaster = weather.WeatherForecaster.from_data(
        average_irradiance,
        solar_irradiance_filenames,
        temperature_filenames,
        os.path.join(location, WEATHER_DATA_FILENAME),
        override_ambient_temperature,
        override_irradiance,
        override_wind_speed,
        use_pvgis,
    )
    return weather_forecaster


######################
# System run methods #
######################


def main(  # pylint: disable=too-many-branches
    average_irradiance: bool,
    cloud_efficacy_factor: float,
    disable_logging: bool,
    exchanger_data_file: str,
    initial_month: int,
    initial_system_temperature_vector: List[float],
    layers: Set[TemperatureName],
    location: str,
    operating_mode: OperatingMode,
    portion_covered: float,
    pvt_data_file: str,
    resolution: int,
    save_2d_output: bool,
    tank_data_file: str,
    use_pvgis: bool,
    verbose: bool,
    x_resolution: int,
    y_resolution: int,
    *,
    run_number: Optional[int],
    start_time: int,
    override_ambient_temperature: Optional[float] = None,
    override_collector_input_temperature: Optional[float] = None,
    override_irradiance: Optional[float] = None,
    override_mass_flow_rate: Optional[float] = None,
    override_wind_speed: Optional[float] = None,
    days: Optional[int] = None,
    minutes: Optional[int] = None,
    months: Optional[int] = None,
) -> Tuple[numpy.ndarray, Dict[float, SystemData]]:
    """
    The main module for the code. Calling this method executes a run of the simulation.

    :param average_irradiance:
        Whether to use an average irradiance profile for the month.

    :param cloud_efficiacy_factor:
        The extent to which cloud cover influences the solar irradiance.

    :param disable_logging:
        Whether to disable the file handler for the logging (True) or keep it enabled
        (False).

    :param exchanger_data_file:
        The path to the data file containing information about the heat exchanger being
        modelled in the run.

    :param initial_month:
        The initial month for the run, given between 1 (January) and 12 (December).

    :param initial_system_temperature_vector:
        The vector of initial temperatures used to model the system.

    :param location:
        The location at which the PVT system is installed and for which location data
        should be used.

    :param operating_mode:
        The operating mode of the run, containing information needed to set up the
        matrix.

    :param portion_covered:
        The portion of the absorber which is covered with PV cells.

    :param pvt_data_file:
        The path to the data file containing information about the PVT system being
        modelled.

    :param resolution:
        The temporal resolution at which to run the simulation.

    :param save_2d_output:
        If True, the 2D output is saved to the system data and returned. If False, only
        the 1D output is saved.

    :param tank_data_file:
        The path to the data file containing information about the hot-water tank used
        in the model.

    :param unglazed:
        If specified, the glass cover is not included in the model.

    :param use_pvgis:
        Whether the data obtained from the PVGIS system should be used.

    :param verbose:
        Whether the logging level is verbose (True) or not (False).

    :param x_resolution:
        The x resolution of the simulation being run.

    :param y_resolution:
        The y resolution of the simulation being run.

    :param run_number:
        The number of the run being carried out. This is used for categorising logs.

    :param start_time:
        The time of day at which to start the simulation, specified between 0 and 23.
        This can be `None` if a steady-state simulation is being run.

    :param override_ambient_temperature:
        In decoupled instances, the ambient temperature can be specified as a constant
        value which will override the ambient-temperature profiles.

    :param override_collector_input_temperature:
        In decoupled instances, the collector input temperature can be specified as a
        constant value which will override the dynamic behaviour. This should be
        specified in Kelvin.

    :param override_irradiance:
        In decoupled instances, the solar irradiance can be specified as a constant
        value which will override the solar-irradiance profiles.

    :param override_mass_flow_rate:
        If provided, this will override the mass-flow rate used in the collector.

    :param override_wind_speed:
        In decoupled instances, the wind speed can be specified as a cosntant value
        which will override the wind-speed profiles.

    :param days:
        The number of days for which the simulation is being run. This can be `None` if
        a steady-state simulation is being run.

    :param minutes:
        The number of minutes for which the simulation is being run. This can be `None`
        if either a steady-state simulation is being run, or if either days or months
        have been specified.

    :param months:
        The number of months for which to run the simulation. This can be `None` if a
        steady-state simulation is being run.

    :return:
        The system data is returned.

    """

    # Get the logger for the component.
    if operating_mode.dynamic:
        logger = get_logger(
            disable_logging,
            PVT_SYSTEM_MODEL_LOGGER_NAME.format(
                tag=f"{resolution}s", run_number=run_number
            ),
            verbose,
        )
    else:
        logger = get_logger(
            disable_logging,
            PVT_SYSTEM_MODEL_LOGGER_NAME.format(
                tag="steady_state", run_number=run_number
            ),
            verbose,
        )

    # Set up numpy printing style.
    numpy.set_printoptions(formatter={"float": "{: 0.3f}".format})

    # Set up the weather module.
    weather_forecaster = _get_weather_forecaster(
        average_irradiance,
        location,
        use_pvgis,
        override_ambient_temperature,
        override_irradiance,
        override_wind_speed,
    )
    logger.info("Weather forecaster successfully instantiated: %s", weather_forecaster)

    # Set up the load module.
    if operating_mode.coupled:
        load_system: Optional[load.LoadSystem] = _get_load_system(location)
        logger.info(
            "Load system successfully instantiated: %s",
            load_system,
        )
    else:
        load_system = None

    # Raise an exception if no initial system temperature vector was supplied.
    if initial_system_temperature_vector is None:
        raise ProgrammerJudgementFault(
            "Not initial system temperature vector was supplied. This is necessary "
            "when calling the pvt model."
        )

    # Initialise the PV-T panel.
    pvt_collector = process_pvt_system_data.pvt_collector_from_path(
        layers,
        logger,
        override_mass_flow_rate,
        portion_covered,
        pvt_data_file,
        x_resolution,
        y_resolution,
    )
    logger.info("PV-T panel successfully instantiated: %s", pvt_collector)
    logger.debug(
        "PV-T panel elements:\n  %s",
        "\n  ".join(
            [
                f"{element_coordinates}: {element}"
                for element_coordinates, element in pvt_collector.elements.items()
            ]
        ),
    )

    # Instantiate the rest of the PVT system if relevant.
    if operating_mode.coupled:
        # Set up the heat exchanger.
        heat_exchanger: Optional[
            exchanger.Exchanger
        ] = process_pvt_system_data.heat_exchanger_from_path(exchanger_data_file)
        logger.info("Heat exchanger successfully instantiated: %s", heat_exchanger)

        # Set up the hot-water tank.
        hot_water_tank: Optional[
            tank.Tank
        ] = process_pvt_system_data.hot_water_tank_from_path(tank_data_file)
        logger.info("Hot-water tank successfully instantiated: %s", hot_water_tank)

        # Set up the mains supply system.
        mains_supply = mains_power.MainsSupply.from_yaml(
            os.path.join(location, "utilities.yaml")
        )
        logger.info("Mains supply successfully instantiated: %s", mains_supply)

    else:
        heat_exchanger = None
        logger.info("NO HEAT EXCHANGER PRESENT")
        hot_water_tank = None
        logger.info("NO HOT-WATER TANK PRESENT")
        mains_supply = None
        logger.info("NO MAINS SUPPLY PRESENT")

    # Determine the number of temperatures being modelled.
    number_of_pipes = len(
        {
            element.pipe_index
            for element in pvt_collector.elements.values()
            if element.pipe_index is not None
        }
    )
    number_of_x_elements = len(
        {element.x_index for element in pvt_collector.elements.values()}
    )
    number_of_y_elements = len(
        {element.y_index for element in pvt_collector.elements.values()}
    )
    if operating_mode.coupled:
        number_of_temperatures: int = index_handler.num_temperatures(pvt_collector)
    else:
        number_of_temperatures = index_handler.num_temperatures(pvt_collector) - 3
    logger.info(
        "System consists of %s pipes, %s by %s elements, and %s temperatures in all.",
        number_of_pipes,
        number_of_x_elements,
        number_of_y_elements,
        number_of_temperatures,
    )

    # Instantiate the two pipes used to store input and output temperature values.
    # absorber_to_tank_pipe = pipe.Pipe(temperature=initial_system_temperature_vector[3])
    # tank_to_absorber_pipe = pipe.Pipe(temperature=initial_system_temperature_vector[3])

    # Instnatiate the hot-water pump.
    # htf_pump = process_pvt_system_data.pump_from_path(pump_data_file)

    # Set up a holder for the information about the final output of the system.
    # total_power_data = TotalPowerData()

    logger.info(
        "System state before beginning run:\n%s\n%s\n%s\n%s",
        heat_exchanger if heat_exchanger is not None else "No heat exchanger",
        hot_water_tank if hot_water_tank is not None else "No hot-water tank",
        pvt_collector,
        weather_forecaster,
    )

    if operating_mode.coupled and operating_mode.dynamic:
        if heat_exchanger is None or hot_water_tank is None or load_system is None:
            raise ProgrammerJudgementFault(
                "{}{} not defined in dynamic operation.{}".format(
                    BColours.FAIL,
                    ", ".join(
                        {
                            entry
                            for entry in {
                                "heat exchanger" if heat_exchanger is None else "",
                                "hot-water tank" if hot_water_tank is None else "",
                                "load system" if load_system is None else "",
                            }
                            if entry is not None
                        }
                    ),
                    BColours.ENDC,
                )
            )
        final_run_temperature_vector, system_data = coupled.coupled_dynamic_run(
            cloud_efficacy_factor,
            days,
            heat_exchanger,
            hot_water_tank,
            initial_month,
            initial_system_temperature_vector,
            load_system,
            logger,
            minutes,
            months,
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            number_of_y_elements,
            operating_mode,
            pvt_collector,
            resolution,
            save_2d_output,
            start_time,
            weather_forecaster,
        )
    elif operating_mode.decoupled and operating_mode.steady_state:
        if override_collector_input_temperature is None:
            raise ProgrammerJudgementFault(
                "{}Override collector input temperature not provided.{}".format(
                    BColours.FAIL, BColours.ENDC
                )
            )
        (
            final_run_temperature_vector,
            system_data_entry,
        ) = decoupled.decoupled_steady_state_run(
            override_collector_input_temperature,
            cloud_efficacy_factor,
            DEFAULT_INITIAL_DATE_AND_TIME.replace(month=initial_month),
            initial_system_temperature_vector,
            logger,
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            number_of_y_elements,
            operating_mode,
            pvt_collector,
            save_2d_output,
            weather_forecaster,
        )
        system_data = {
            override_collector_input_temperature
            - ZERO_CELCIUS_OFFSET: system_data_entry[1]
        }
    elif operating_mode.decoupled and operating_mode.dynamic:
        if override_collector_input_temperature is None:
            raise ProgrammerJudgementFault(
                "{}Override collector input temperature not provided.{}".format(
                    BColours.FAIL, BColours.ENDC
                )
            )
        (final_run_temperature_vector, system_data,) = decoupled.decoupled_dynamic_run(
            cloud_efficacy_factor,
            override_collector_input_temperature,
            days,
            initial_month,
            initial_system_temperature_vector,
            logger,
            minutes,
            months,
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            number_of_y_elements,
            operating_mode,
            pvt_collector,
            resolution,
            save_2d_output,
            start_time,
            weather_forecaster,
        )
    else:
        raise ProgrammerJudgementFault(
            "The system model was called with an invalid operating mode: {}.".format(
                operating_mode
            )
        )

    return final_run_temperature_vector, system_data


if __name__ == "__main__":
    logging.error(
        "Calling the internal `pvt_system` from the command-line is no longer "
        "supported."
    )
    raise ProgrammerJudgementFault(
        "Calling the internal `pvt_system` from the command-line interface is no "
        "longer supported."
    )



#!/usr/bin/python3.7
########################################################################################
# load.py - Computs load characteristics.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The load module for this PV-T model.

This module computes the load at any given instance based on various probabilistic
factors read in from a data file and stored within the module when running.

"""

import collections
import csv
import datetime
import enum

import json

from dataclasses import dataclass
from typing import Any, Dict, Set, Tuple

from .__utils__ import (
    MissingDataError,
    ProgrammerJudgementFault,
    BaseDailyProfile,
)

__all__ = (
    "LoadData",
    "LoadProfile",
    "LoadSystem",
    "ProfileType",
)


################
# Data Classes #
################


class ProfileType(enum.Enum):
    """
    Used to label whether a profile is a hot-water or electrical (or other) profile.

    """

    ELECTRICITY = 0
    HOT_WATER = 1


@dataclass
class LoadData:
    """
    Conaints information about the various loads on the system.

    .. attribute:: electrical_load
        The electrical load on the system, measured in Watts.

    .. attribute:: hot_water_load
        The hot-water load on the system, measured in Litres.

    """

    electrical_load: float
    hot_water_load: float


class _DayType(enum.Enum):
    """
    Contains information about the type of day.

    .. attribute:: WEEKDAY
        A weekday is being represented here.

    .. attribute:: WEEKEND
        A weekend day is being represented here.

    .. attribute:: DAYLESS
        The type of day, i.e., weekday or weekend, is irrelevant. This information is
        carried around the code as a :class:`_DayType`.DAYLESS instance.

    """

    DAYLESS = 0
    WEEKDAY = 1
    WEEKEND = 2

    @classmethod
    def from_day_number(cls, day_number: int) -> Any:
        """
        Return a :class:`_Day` instance based on the day of the week being processed.

        :param day_number:
            The day number, ie mon = 1, tue = 2, etc..

        :return:
            A :class:`_Day` instance based on this.

        """

        if not isinstance(day_number, int):
            raise Exception("The day must be an integer!")

        if 0 <= day_number <= 4:
            return cls(1)
        if 5 <= day_number <= 6:
            return cls(2)
        if day_number == 7:
            return cls(0)
        raise Exception("The day must fall between 0 and 6 inclusive, or 7 for no day.")


@dataclass
class _MonthAndDayType:
    """
    Contains information about the month and the weekday.

    .. attribute:: month
        The month, expressed as an integer.

    .. attribute:: day_type
        The day type, expressed as a :class:`_DayType`.

    """

    month: int
    day_type: _DayType

    def __hash__(self) -> int:
        """
        Generates a hash s.t. :class:`_MonthAndDayType` instances can be used as keys.

        :return:
            A hash of the :class:`_MonthAndDayType` instnace.

        """

        return hash(self.month * 31 + self.day_type.value)

    @classmethod
    def from_date(cls, date: datetime.date) -> Any:
        """
        Returns a :class:`_MonthAndWeekday` instance based on the date.

        :param date:
            The date for which to generate a month and weekday instance, passed in as a
            :class:`datetime.date` instance.

        :return:
            An instantiated :class:`_MonthAndWeekday` instance.

        """

        return cls(date.month, _DayType.from_day_number(date.weekday()))


class LoadProfile(BaseDailyProfile):
    """
    Represents a load profile.

    .. attribute:: resolution
        The time-step resolution of the profile, measured in seconds.

    """

    # Private Attributes:
    #
    #  .. attribute:: _profile
    #   A mapping of hour to the load at that hour.
    #

    def __init__(
        self,
        profile: Dict[datetime.time, float] = None,
        resolution=1,
    ) -> None:
        """
        Instantiate a load profile.

        :param resolution:
            The resolution of the profile, measured in seconds.

        :param profile:
            The profile data: a mapping of time to the load at that time step.

        """

        super().__init__(profile)

        self.resolution = resolution


#############################
# Internal Helper Functions #
#############################


def _process_csv(
    csv_file_path: str,
) -> Tuple[ProfileType, Dict[_MonthAndDayType, LoadProfile]]:
    """
    Processs a CSV data file and return a mapping of month and day to profile.

    :param csv_file_path:
        The path to the csv data file.

    :return:
        A `tuple` containing the profile type and a  mapping between the
        :class:`_MonthAndDayType` and the load profile as a dict.

    """

    # @@@ For now, the fact that the CSV data is an electricity profile is hard-
    # @@@ -coded in. Ideally, there would be some regex matching here.
    profile_type = ProfileType.ELECTRICITY

    electrical_load_profile: Dict[
        _MonthAndDayType, LoadProfile
    ] = collections.defaultdict(LoadProfile)

    # Open the csv file, and cycle through the rows, processing them.
    with open(csv_file_path, "r") as csv_file:
        file_reader = csv.reader(csv_file, delimiter=",")

        # Cycle through the rows, and process them accordingly.
        for index, row in enumerate(file_reader):
            # Skip the title rows
            if index < 2:
                continue

            # Determine the current time.
            time = datetime.datetime.strptime(row[0], "%H:%M").time()

            # Set the various profile values accordingly.
            electrical_load_profile[_MonthAndDayType(1, _DayType(1))].update(
                {time: float(row[2])}
            )
            electrical_load_profile[_MonthAndDayType(1, _DayType(2))].update(
                {time: float(row[6])}
            )
            electrical_load_profile[_MonthAndDayType(2, _DayType(1))].update(
                {time: float(row[10])}
            )
            electrical_load_profile[_MonthAndDayType(2, _DayType(2))].update(
                {time: float(row[14])}
            )
            electrical_load_profile[_MonthAndDayType(3, _DayType(1))].update(
                {time: float(row[18])}
            )
            electrical_load_profile[_MonthAndDayType(3, _DayType(2))].update(
                {time: float(row[22])}
            )
            electrical_load_profile[_MonthAndDayType(4, _DayType(1))].update(
                {time: float(row[26])}
            )
            electrical_load_profile[_MonthAndDayType(4, _DayType(2))].update(
                {time: float(row[30])}
            )
            electrical_load_profile[_MonthAndDayType(5, _DayType(1))].update(
                {time: float(row[34])}
            )
            electrical_load_profile[_MonthAndDayType(5, _DayType(2))].update(
                {time: float(row[38])}
            )
            electrical_load_profile[_MonthAndDayType(6, _DayType(1))].update(
                {time: float(row[42])}
            )
            electrical_load_profile[_MonthAndDayType(6, _DayType(2))].update(
                {time: float(row[46])}
            )
            electrical_load_profile[_MonthAndDayType(7, _DayType(1))].update(
                {time: float(row[50])}
            )
            electrical_load_profile[_MonthAndDayType(7, _DayType(2))].update(
                {time: float(row[54])}
            )
            electrical_load_profile[_MonthAndDayType(8, _DayType(1))].update(
                {time: float(row[58])}
            )
            electrical_load_profile[_MonthAndDayType(8, _DayType(2))].update(
                {time: float(row[62])}
            )
            electrical_load_profile[_MonthAndDayType(9, _DayType(1))].update(
                {time: float(row[66])}
            )
            electrical_load_profile[_MonthAndDayType(9, _DayType(2))].update(
                {time: float(row[70])}
            )
            electrical_load_profile[_MonthAndDayType(10, _DayType(1))].update(
                {time: float(row[74])}
            )
            electrical_load_profile[_MonthAndDayType(10, _DayType(2))].update(
                {time: float(row[78])}
            )
            electrical_load_profile[_MonthAndDayType(11, _DayType(1))].update(
                {time: float(row[82])}
            )
            electrical_load_profile[_MonthAndDayType(11, _DayType(2))].update(
                {time: float(row[86])}
            )
            electrical_load_profile[_MonthAndDayType(12, _DayType(1))].update(
                {time: float(row[90])}
            )
            electrical_load_profile[_MonthAndDayType(12, _DayType(2))].update(
                {time: float(row[94])}
            )

    return profile_type, electrical_load_profile


def _process_json(
    json_file_path: str,
) -> Tuple[ProfileType, Dict[_MonthAndDayType, LoadProfile]]:
    """
    Processs a JSON data file and return a mapping of month and day to profile.

    :param json_file_path:
        The path to the json data file.

    :return:
        A `tuple` containing the profile type and a  mapping between the
        :class:`_MonthAndDayType` and the load profile as a dict.

    """

    if "thermal" in json_file_path:
        profile_type = ProfileType.HOT_WATER
    elif "electrical" in json_file_path:
        profile_type = ProfileType.ELECTRICITY
    else:
        raise ProgrammerJudgementFault(
            "The profle type supplied is not electrical or thermal. See load module."
        )

    # @@@ For now, the data is stored irrespective of month and day. This is in order to
    # @@@ match as closely as possible to Maria's data set for July.
    load_profile: Dict[_MonthAndDayType, LoadProfile] = collections.defaultdict(
        LoadProfile
    )

    with open(json_file_path, "r") as f:
        json_data = json.load(f)

    # Process the data.
    try:
        load_profile[_MonthAndDayType(0, _DayType(0))] = LoadProfile(
            {
                datetime.datetime.strptime(key, "%H:%M:%S.%f").time(): float(value)
                for key, value in json_data.items()
            }
        )
        return profile_type, load_profile
    except ValueError:
        pass

    # Attempt using a different time format.
    try:
        load_profile[_MonthAndDayType(0, _DayType(0))] = LoadProfile(
            {
                datetime.datetime.strptime(key, "%H:%M:%S").time(): float(value)
                for key, value in json_data.items()
            }
        )
        return profile_type, load_profile
    except ValueError:
        pass

    # Attempt using a different time format.
    try:
        load_profile[_MonthAndDayType(0, _DayType(0))] = LoadProfile(
            {
                datetime.datetime.strptime(key, "%H:%M").time(): float(value)
                for key, value in json_data.items()
            }
        )
        return profile_type, load_profile
    except ValueError as e:
        raise ProgrammerJudgementFault(
            f"The load profile data is of an incompatible time format: {str(e)}"
        ) from None


#########################
# External Load Classes #
#########################


class LoadSystem:
    """
    Represents a system, eg, a household, potential of generating loads.

    """

    # Private Attributes:
    #
    # .. attribute:: _monthly_load_profiles
    #   A mapping between the profile type, month, and day of the profile.
    #

    def __init__(
        self,
        seasonal_load_profiles: Dict[ProfileType, Dict[_MonthAndDayType, LoadProfile]],
    ) -> None:
        """
        Instantiate a Load System.

        :param seasonal_load_profiles:
            A mapping representing the internal structure of how load profiles are to be
            stored.

        """

        self._seasonal_load_profiles = seasonal_load_profiles

    def __getitem__(self, index: Tuple[ProfileType, datetime.datetime]) -> float:
        """
        Returns the load value based on the profile type wanted and the date and time.

        :param index:
            A `tuple` containing the profile type wanted and the date and time for which
            to fetch the value.

        :return:
            The value of the given profile at the given date and time.

        """

        profile_type, date_and_time = index
        month_and_day = _MonthAndDayType.from_date(date_and_time.date())

        try:
            return self._seasonal_load_profiles[profile_type][month_and_day][
                date_and_time.time()
            ]
        except (KeyError, ValueError):
            pass

        # * Attempt to get dayless data.
        month_and_day.day_type = _DayType(0)
        try:
            return self._seasonal_load_profiles[profile_type][month_and_day][
                date_and_time.time()
            ]
        except (KeyError, ValueError):
            pass

        # * Attempt to get seasonless data
        month_and_day.month = 0
        try:
            return self._seasonal_load_profiles[profile_type][month_and_day][
                date_and_time.time()
            ]
        except (KeyError, ValueError) as e:
            raise MissingDataError(
                "Could not find profile data for given key. Key: {}. Error: {}".format(
                    str(index),
                    str(e),
                )
            ) from None

    @classmethod
    def from_data(cls, data_file_paths: Set[str]) -> Any:  # type: ignore
        """
        Returns a :class:`LoadSystem` based on the paths to various data files.

        :param data_file_paths:
            A set of paths to various data files.

        :return:
            A :class:`LoadSystem` instance, instantiated from the passed in data files.

        """

        seasonal_load_profiles: Dict[
            ProfileType, Dict[_MonthAndDayType, LoadProfile]
        ] = {
            ProfileType.ELECTRICITY: collections.defaultdict(LoadProfile),
            ProfileType.HOT_WATER: collections.defaultdict(LoadProfile),
        }

        for data_file_name in data_file_paths:
            if data_file_name.endswith(".csv"):
                profile_type, profile_data = _process_csv(data_file_name)
            elif data_file_name.endswith(".json"):
                profile_type, profile_data = _process_json(data_file_name)
            else:
                raise ProgrammerJudgementFault(
                    "Only .csv and .json files supported for load profiles. "
                    f"Filename attempted: {data_file_name}"
                )
            seasonal_load_profiles[profile_type].update(profile_data)

        return cls(seasonal_load_profiles)



#!/usr/bin/python3.7
########################################################################################
# exchanger.py - The exchanger module for this, my first, PV-T model! :O
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The exchanger module for the PV-T model.

This module represents the heat exchanger within the hot-water tank.

"""

__all__ = ("Exchanger",)


class Exchanger:
    """
    Represents a physical heat exchanger within a hot-water tank.

    .. attribute:: efficiency
        The efficiency of the heat exchanger, defined between 0 and 1.

    """

    def __init__(self, efficiency) -> None:
        """
        Instantiate a heat exchanger instance.

        :param efficiency:
            The efficiency of the heat exchanger, defined between 0 and 1.

        """

        self.efficiency = efficiency

    def __repr__(self) -> str:
        """
        Returns a nice representation of the heat exchanger.

        :return:
            A `str` giving a nice representation of the heat exchanger.

        """

        return f"Exchanger(efficiency: {self.efficiency})"

    def get_heat_addition(
        self,
        input_water_mass_flow_rate: float,
        input_water_heat_capacity: float,
        input_water_temperature: float,
        water_tank_temperature: float,
    ) -> float:
        """
        Computes the heat added to the hot-water tank.

        :param input_water_heat_capacity:
            The heat capacity of the HTF, measured in Joules per kilogram Kelvin.

        :param input_water_mass_flow_rate:
            The mass flow rate of water through the HTF side of the heat exchanger,
            measured in kilograms per second.

        :param input_water_temperature:
            The tempertaure of the HTF being inputted on the HTF side of the heat
            exchanger.

        :param water_tank_temperature:
            The hot-water tank temperature, measured in Kelvin.

        :return:
            The heat addition to the hot-water tank from the exchanger, measured in
            Watts.

        """

        tank_heat_addition: float = (
            self.efficiency
            * input_water_mass_flow_rate  # [kg/s]
            * input_water_heat_capacity  # [J/kg*K]
        ) * (
            input_water_temperature - water_tank_temperature  # [K]
        )  # [W]

        return tank_heat_addition

    def get_output_htf_temperature(
        self,
        input_water_temperature: float,
        water_tank_temperature: float,
    ) -> float:
        """
        Computes the temperature of the HTF leaving the heat exchanger, in Kelvin.

        :param input_water_temperature:
            The temperature of the HTF entering the heat exchanger, measured in Kelvin.

        :param water_tank_temperature:
            The temperature of the hot-water tank, measured in Kelvin.

        :return:
            The temperature of the HTF leaving the heat exchanger, measured in Kelvin.

        """

        exchanger_output_temperature: float = (
            input_water_temperature
            - self.efficiency * (input_water_temperature - water_tank_temperature)
        )

        return exchanger_output_temperature



#!/usr/bin/python3.7
########################################################################################
# glass.py - The glass module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The glass module for the matrix component.

This module computes and returns the equation(s) associated with the glass layer of the
PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""


import logging

from typing import List, Optional, Tuple, Union

import numpy

from .. import index_handler, physics_utils
from ..pvt_collector import pvt

from ...__utils__ import (
    BColours,
    OperatingMode,
    ProgrammerJudgementFault,
    TemperatureName,
)
from ..__utils__ import WeatherConditions
from ..pvt_collector.element import Element, ElementCoordinates


__all__ = ("calculate_glass_equation",)


def calculate_glass_equation(  # pylint: disable=too-many-branches
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    glass_downward_conduction: float,
    glass_downward_radiation: float,
    logger: logging.Logger,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    previous_temperature_vector: Optional[numpy.ndarray],
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    element: Element,
    upper_glass_downward_conduction: float,
    upper_glass_downward_radiation: float,
    weather_conditions: WeatherConditions,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the glass equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    logger.debug(
        "Beginning calculation of glass equation for element %s.", element.coordinates
    )

    if pvt_collector.glass is None:
        raise ProgrammerJudgementFault(
            "{}Element {} has a glass layer but no glass data supplied ".format(
                BColours.FAIL, element
            )
            + "in the PV-T data file.{}".format(BColours.ENDC)
        )

    if operating_mode.dynamic:
        glass_internal_energy: float = (
            element.width  # [m]
            * element.length  # [m]
            * pvt_collector.glass.thickness  # [m]
            * pvt_collector.glass.density  # [kg/m^3]
            * pvt_collector.glass.heat_capacity  # [J/kg*K]
            / resolution  # type: ignore  # [s]
        )
    else:
        glass_internal_energy = 0
    logger.debug("Glass internal energy term: %s W/K", glass_internal_energy)

    # Compute the positive conductive term based on the next element along.
    positive_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index + 1, element.y_index)
    )
    if positive_x_element is not None:
        positive_x_wise_conduction: float = (
            pvt_collector.glass.conductivity  # [W/m*K]
            * pvt_collector.glass.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + positive_x_element.width))  # [m]
        )
    else:
        positive_x_wise_conduction = 0
    logger.debug(
        "Positive glass x-wise conduction term: %s W/K", positive_x_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index - 1, element.y_index)
    )
    if negative_x_element is not None:
        negative_x_wise_conduction: float = (
            pvt_collector.glass.conductivity  # [W/m*K]
            * pvt_collector.glass.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + negative_x_element.width))  # [m]
        )
    else:
        negative_x_wise_conduction = 0
    logger.debug(
        "Negative glass x-wise conduction term: %s W/K", negative_x_wise_conduction
    )

    # Compute the overall x-wise conduction term.
    x_wise_conduction = positive_x_wise_conduction + negative_x_wise_conduction
    logger.debug("Glass x-wise conduction term: %s W/K", x_wise_conduction)

    # Compute the positive conductive term based on the next element along.
    positive_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index + 1)
    )
    if positive_y_element is not None:
        positive_y_wise_conduction: float = (
            pvt_collector.glass.conductivity  # [W/m*K]
            * pvt_collector.glass.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + positive_y_element.length))  # [m]
        )
    else:
        positive_y_wise_conduction = 0
    logger.debug(
        "Positive glass y-wise conduction term: %s W/K", positive_y_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index - 1)
    )
    if negative_y_element is not None:
        negative_y_wise_conduction: float = (
            pvt_collector.glass.conductivity  # [W/m*K]
            * pvt_collector.glass.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + negative_y_element.length))  # [m]
        )
    else:
        negative_y_wise_conduction = 0
    logger.debug(
        "Negative glass y-wise conduction term: %s W/K", negative_y_wise_conduction
    )

    # Compute the overall y-wise conduction term.
    y_wise_conduction = positive_y_wise_conduction + negative_y_wise_conduction
    logger.debug("Glass y-wise conduction term: %s W/K", y_wise_conduction)

    if not element.upper_glass:
        (
            glass_to_air_conduction,
            glass_to_sky_radiation,
        ) = physics_utils.upward_loss_terms(
            best_guess_temperature_vector,
            pvt_collector,
            element,
            pvt_collector.glass.emissivity,
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index,
                element.y_index,
            ),
            weather_conditions,
        )
        logger.debug("Glass to air conduction %s W/K", glass_to_air_conduction)
        logger.debug("Glass to sky radiation %s W/K", glass_to_sky_radiation)
    else:
        glass_to_air_conduction = 0
        logger.debug("Glass layer does not conduct to the air due to double glazing.")
        glass_to_sky_radiation = 0
        logger.debug("Glass layer does not radiate to the sky due to double glazing.")

    # Compute the T_g(i, j) term
    row_equation[
        index_handler.index_from_element_coordinates(
            number_of_x_elements,
            pvt_collector,
            TemperatureName.glass,
            element.x_index,
            element.y_index,
        )
    ] = (
        glass_internal_energy
        + x_wise_conduction
        + y_wise_conduction
        + glass_to_air_conduction
        + glass_to_sky_radiation
        + glass_downward_conduction
        + glass_downward_radiation
        + upper_glass_downward_conduction
        + upper_glass_downward_radiation
    )

    # Compute the T_g(i+1, j) term provided that that element exists.
    if element.x_index + 1 < number_of_x_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index + 1,
                element.y_index,
            )
        ] = (
            -1 * positive_x_wise_conduction
        )

    # Compute the T_g(i-1, j) term provided that that element exists.
    if element.x_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index - 1,
                element.y_index,
            )
        ] = (
            -1 * negative_x_wise_conduction
        )

    # Compute the T_g(i, j+1) term provided that that element exists.
    if element.y_index + 1 < number_of_y_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index,
                element.y_index + 1,
            )
        ] = (
            -1 * positive_y_wise_conduction
        )

    # Compute the T_g(i, j-1) term provided that that element exists.
    if element.y_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index,
                element.y_index - 1,
            )
        ] = (
            -1 * negative_y_wise_conduction
        )

    # Compute the T_ug(i, j) term provided that there is an upper-glass layer present.
    if element.upper_glass:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.upper_glass,
                element.x_index,
                element.y_index,
            )
        ] = -1 * (upper_glass_downward_conduction + upper_glass_downward_radiation)

    # Compute the T_pv(i, j) term provided that there is a PV layer present.
    if element.pv:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index,
                element.y_index,
            )
        ] = -1 * (glass_downward_conduction + glass_downward_radiation)
    # Otherwise, compute the T_A(i, j) term.
    else:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index,
                element.y_index,
            )
        ] = -1 * (glass_downward_conduction + glass_downward_radiation)

    # Compute the resultant vector value.
    resultant_vector_value = (
        # Ambient temperature term.
        glass_to_air_conduction * weather_conditions.ambient_temperature  # [W]
        # Sky temperature term.
        + glass_to_sky_radiation * weather_conditions.sky_temperature  # [W]
        # Solar absorption term.
        + element.width  # [m]
        * element.length  # [m]
        * pvt_collector.glass_transmissivity_absorptivity_product
        * weather_conditions.irradiance  # [W/m^2]
    )

    # if len(numpy.argwhere(numpy.isnan(row_equation))) > 0:
    #     import pdb

    #     pdb.set_trace()

    if operating_mode.dynamic:
        resultant_vector_value += (
            # Previous glass temperature term.
            glass_internal_energy  # [W/K]
            * previous_temperature_vector[  # type: ignore
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.glass,
                    element.x_index,
                    element.y_index,
                )
            ]  # [K]
        )

        logger.debug(
            "Rough Glass Temperature estimate: %s K.",
            int(
                resultant_vector_value
                / (
                    glass_to_air_conduction
                    + glass_to_sky_radiation
                    + glass_internal_energy
                )
            ),
        )

    return row_equation, resultant_vector_value



#!/usr/bin/python3.7
########################################################################################
# convergent_solver.py - Module for recursively calculating the matrix until convergence
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
Module for recursively calculating the matrix until convergence

This module recursively calls the matrix module until convergence is reached.

"""

import datetime
import logging

from typing import List, Optional, Union

import numpy

from scipy import linalg  # type: ignore

from . import exchanger, tank
from .matrix import matrix
from .pvt_collector import pvt

from ..__utils__ import BColours, OperatingMode, ProgrammerJudgementFault
from .__utils__ import (
    DivergentSolutionError,
    PVT_SYSTEM_MODEL_LOGGER_NAME,
    WeatherConditions,
)
from .constants import (
    CONVERGENT_SOLUTION_PRECISION,
    MAXIMUM_RECURSION_DEPTH,
    WARN_RECURSION_DEPTH,
    ZERO_CELCIUS_OFFSET,
)

__all__ = ("solve_temperature_vector_convergence_method",)


def _calculate_vector_difference(
    first_vector: Union[List[float], numpy.ndarray],
    second_vector: Union[List[float], numpy.ndarray],
) -> float:
    """
    Computes a measure of the difference between two vectors.

    :param first_vector:
        The first vector.

    :param second_vector:
        The second vector.

    :return:
        A measure of the difference between the two vectors.

    """

    # Compute the gross difference between the vectors.
    try:
        diff_vector: List[float] = [
            first_vector[index] - second_vector[index]
            for index in range(len(first_vector))
        ]
    except ValueError as e:
        raise ProgrammerJudgementFault(
            "Atempt was made to compute the difference between two vectors of "
            f"different sizes: {str(e)}"
        ) from None

    # Take the sum of the values in this vector to avoid sign issues and return the sum.
    return sum({abs(value) for value in diff_vector})


def solve_temperature_vector_convergence_method(
    *,
    logger: logging.Logger,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    run_one_temperature_vector: Union[List[float], numpy.ndarray],
    weather_conditions: WeatherConditions,
    convergence_run_number: int = 0,
    run_one_temperature_difference: float = 5 * ZERO_CELCIUS_OFFSET ** 2,
    collector_input_temperature: Optional[float] = None,
    current_hot_water_load: Optional[float] = None,
    heat_exchanger: Optional[exchanger.Exchanger] = None,
    hot_water_tank: Optional[tank.Tank] = None,
    next_date_and_time: Optional[datetime.datetime] = None,
    previous_run_temperature_vector: Optional[Union[List[float], numpy.ndarray]] = None,
    resolution: Optional[int] = None,
) -> Union[List[float], numpy.ndarray]:
    """
    Itteratively solves for the temperature vector to find a convergent solution.

    The method used to compute the temperatures at the next time step involves
    approximating a non-linear temperature dependance using a best-guess set of
    temperatures for the temperatures at the next time step.

    The "best guess" temperature vector is fed into the matrix solver method, and, from
    this, approximations for the non-linear terms are computed. This then returns a
    "best guess" matrix, from which the temperatures at the next time step are computed.

    Whether the solution is converging, or has converged, is determined.
        - If the solution has converged, the temperature vector computed is returned;
        - If the solution has diverged, an error is raised;
        - If the solution is converging, but has not yet converged, then the function
          runs through again.

    :param collector_input_temperature:
        If set, this governs the input water temperature of the feedwater entering the
        absorber.

    :param current_hot_water_load:
        The current hot-water load placed on the system, measured in kilograms per
        second.

    :param heat_exchanger:
        The heat exchanger being modelled in the system.

    :param hot_water_tank:
        A :class:`tank.Tank` instance representing the hot-water tank being modelled in
        the system.

    :param logger:
        The logger for the module run.

    :param next_date_and_time:
        The date and time at the time step being solved.

    :param number_of_pipes:
        The nmber of pipes on the PVT panel.

    :param number_of_temperatures:
        The number of unique temperature values to determine for the model.

    :param number_of_x_elements:
        The number of elements in one "row" of the panel, i.e., which have the same y
        coordinate

    :param number_of_y_elements:
        The number of elements in one "column" of the panel, i.e., which have the same x
        coordiante.

    :param operating_mode:
        The operating mode for the run.

    :param previous_run_temperature_vector:
        The temperatures at the previous time step.

    :param pvt_collector:
        A :class:`pvt.PVT` instance representing the PVT panel being modelled.

    :param resolution:
        The resolution used for the run, measured in seconds.

    :param run_one_temperature_vector:
        The temperature vector at the last run of the convergent solver.

    :param weather_conditions:
        The weather conditions at the time step being computed.

    :param run_one_temperature_difference:
        The temperature difference between the two vectors when the function was
        previously run.

    :return:
        The temperatures at the next time step.

    :raises: DivergentSolutionError
        Raised if the solution starts to diverge.

    """

    logger.info(
        "Date and time: %s; Run number: %s: Beginning convergent calculation.",
        next_date_and_time.strftime("%d/%m/%Y %H:%M:%S")
        if next_date_and_time is not None
        else "[time-independent]",
        convergence_run_number,
    )

    if operating_mode.coupled:
        coefficient_matrix, resultant_vector = matrix.calculate_matrix_equation(
            best_guess_temperature_vector=run_one_temperature_vector,
            heat_exchanger=heat_exchanger,
            hot_water_load=current_hot_water_load,
            hot_water_tank=hot_water_tank,
            logger=logger,
            number_of_pipes=number_of_pipes,
            number_of_temperatures=number_of_temperatures,
            number_of_x_elements=number_of_x_elements,
            number_of_y_elements=number_of_y_elements,
            operating_mode=operating_mode,
            previous_temperature_vector=numpy.asarray(previous_run_temperature_vector),
            pvt_collector=pvt_collector,
            resolution=resolution,
            weather_conditions=weather_conditions,
        )
    elif operating_mode.steady_state and operating_mode.decoupled:
        coefficient_matrix, resultant_vector = matrix.calculate_matrix_equation(
            best_guess_temperature_vector=run_one_temperature_vector,
            collector_input_temperature=collector_input_temperature,
            logger=logger,
            number_of_pipes=number_of_pipes,
            number_of_temperatures=number_of_temperatures,
            number_of_x_elements=number_of_x_elements,
            number_of_y_elements=number_of_y_elements,
            operating_mode=operating_mode,
            pvt_collector=pvt_collector,
            resolution=resolution,
            weather_conditions=weather_conditions,
        )
    elif operating_mode.dynamic and operating_mode.decoupled:
        coefficient_matrix, resultant_vector = matrix.calculate_matrix_equation(
            best_guess_temperature_vector=run_one_temperature_vector,
            collector_input_temperature=collector_input_temperature,
            logger=logger,
            number_of_pipes=number_of_pipes,
            number_of_temperatures=number_of_temperatures,
            number_of_x_elements=number_of_x_elements,
            number_of_y_elements=number_of_y_elements,
            operating_mode=operating_mode,
            previous_temperature_vector=numpy.asarray(previous_run_temperature_vector),
            pvt_collector=pvt_collector,
            resolution=resolution,
            weather_conditions=weather_conditions,
        )

    logger.debug(
        "Matrix equation computed.\nA =\n%s\nB =\n%s",
        str("\n".join([",".join([str(e) for e in row]) for row in coefficient_matrix])),
        str(resultant_vector),
    )

    try:
        run_two_output = linalg.solve(a=coefficient_matrix, b=resultant_vector)
    except ValueError:
        logger.error(
            "%sMatrix has dimensions %s and is not square.%s",
            BColours.FAIL,
            coefficient_matrix.shape,
            BColours.ENDC,
        )
        raise
    # run_two_output = linalg.lstsq(a=coefficient_matrix, b=resultant_vector)
    run_two_temperature_vector: Union[List[float], numpy.ndarray] = numpy.asarray(  # type: ignore
        [run_two_output[index][0] for index in range(len(run_two_output))]
    )
    # run_two_temperature_vector = run_two_output[0].transpose()[0]

    logger.debug(
        "Date and time: %s; Run number: %s: "
        "Temperatures successfully computed. Temperature vector: T = %s",
        next_date_and_time.strftime("%d/%m/%Y %H:%M:%S")
        if next_date_and_time is not None
        else "[time-independent]",
        convergence_run_number,
        run_two_temperature_vector,
    )

    run_two_temperature_difference = _calculate_vector_difference(
        run_one_temperature_vector, run_two_temperature_vector
    )

    # If the solution has converged, return the temperature vector.
    if run_two_temperature_difference < (10 ** -CONVERGENT_SOLUTION_PRECISION):
        logger.debug(
            "Date and time: %s; Run number: %s: Convergent solution found. "
            "Convergent difference: %s",
            next_date_and_time.strftime("%d/%m/%Y %H:%M:%S")
            if next_date_and_time is not None
            else "[time-independent]",
            convergence_run_number,
            run_two_temperature_difference,
        )
        return run_two_temperature_vector

    logger.debug(
        "Date and time: %s; Run number: %s: Solution not convergent. "
        "Convergent difference: %s",
        next_date_and_time.strftime("%d/%m/%Y %H:%M:%S")
        if next_date_and_time is not None
        else "[time-independent]",
        convergence_run_number,
        run_two_temperature_difference,
    )

    if convergence_run_number >= WARN_RECURSION_DEPTH:
        logger.warn(
            "%sRun depth greater than, or equal, to 10 reached when attempting a "
            "convergent solution.%s",
            BColours.FAIL,
            BColours.ENDC,
        )

    if convergence_run_number >= MAXIMUM_RECURSION_DEPTH:
        logger.error(
            "%sRun depth exceeded maximum recursion depth of %s reached when "
            "attempting a convergent solution.%s",
            BColours.FAIL,
            MAXIMUM_RECURSION_DEPTH,
            BColours.ENDC,
        )
        raise RecursionError(
            BColours.FAIL
            + "Maximum recursion level specified by the constants module reached."
            + BColours.ENDC
        )

    # If the solution has diverged, raise an Exception.
    if round(run_two_temperature_difference, CONVERGENT_SOLUTION_PRECISION) > round(
        run_one_temperature_difference, CONVERGENT_SOLUTION_PRECISION
    ):
        if convergence_run_number > 2:
            logger.error(
                "The temperature solutions at the next time step diverged. "
                "See %s for more details.",
                PVT_SYSTEM_MODEL_LOGGER_NAME,
            )
            logger.info(
                "Local variables at the time of the dump:\n%s",
                "\n".join([f"{key}: {value}" for key, value in locals().items()]),
            )
            raise DivergentSolutionError(
                convergence_run_number,
                run_one_temperature_difference,
                run_one_temperature_vector,
                run_two_temperature_difference,
                run_two_temperature_vector,
            )
        logger.debug("Continuing as fewer than two runs have been attempted...")

    # Otherwise, continue to solve until the prevision is reached.
    return solve_temperature_vector_convergence_method(
        collector_input_temperature=collector_input_temperature,
        current_hot_water_load=current_hot_water_load,
        heat_exchanger=heat_exchanger,
        hot_water_tank=hot_water_tank,
        logger=logger,
        next_date_and_time=next_date_and_time,
        number_of_pipes=number_of_pipes,
        number_of_temperatures=number_of_temperatures,
        number_of_x_elements=number_of_x_elements,
        number_of_y_elements=number_of_y_elements,
        operating_mode=operating_mode,
        previous_run_temperature_vector=previous_run_temperature_vector,
        pvt_collector=pvt_collector,
        resolution=resolution,
        run_one_temperature_vector=run_two_temperature_vector,
        weather_conditions=weather_conditions,
        convergence_run_number=convergence_run_number + 1,
        run_one_temperature_difference=run_two_temperature_difference,
    )



#!/usr/bin/python3.7
########################################################################################
# pv.py - The pv module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The pv module for the matrix component.

This module computes and returns the equation(s) associated with the PV layer of the
PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""

import logging

from typing import List, Optional, Tuple, Union

import numpy

from .. import index_handler, physics_utils
from ..pvt_collector import pvt

from ...__utils__ import OperatingMode, TemperatureName
from ..__utils__ import WeatherConditions
from ..efficiency import electrical_efficiency
from ..pvt_collector.element import Element, ElementCoordinates

__all__ = ("calculate_pv_equation",)


def calculate_pv_equation(  # pylint: disable=too-many-branches
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    logger: logging.Logger,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    previous_temperature_vector: Optional[numpy.ndarray],
    pv_to_absorber_conduction: float,
    pv_to_glass_conduction: float,
    pv_to_glass_radiation: float,
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    element: Element,
    weather_conditions: WeatherConditions,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the pv equation.

    :param logger:
        The logger used in the run.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    # if weather_conditions.irradiance > 0:
    #     import pdb

    #     pdb.set_trace(header=f"T_PV{element.coordinates}")

    logger.debug(
        "Beginning calculation of PV equation for element %s.", element.coordinates
    )

    if operating_mode.dynamic:
        pv_internal_energy: float = (
            element.width  # [m]
            * element.length  # [m]
            * pvt_collector.pv.thickness  # [m]
            * pvt_collector.pv.density  # [kg/m^3]
            * pvt_collector.pv.heat_capacity  # [J/kg*K]
            / resolution  # type: ignore  # [s]
        )
    else:
        pv_internal_energy = 0
    logger.debug("PV internal energy term: %s W/K", pv_internal_energy)

    # Compute the upward environmental losses.
    if not element.glass:
        (pv_to_air_conduction, pv_to_sky_radiation,) = physics_utils.upward_loss_terms(
            best_guess_temperature_vector,
            pvt_collector,
            element,
            pvt_collector.pv.emissivity,
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index,
                element.y_index,
            ),
            weather_conditions,
        )
        logger.debug("PV to air conduction %s W/K", pv_to_air_conduction)
        logger.debug("PV to sky radiation %s W/K", pv_to_sky_radiation)
    else:
        pv_to_air_conduction = 0
        pv_to_sky_radiation = 0

    # Compute the positive conductive term based on the next element along.
    positive_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index + 1, element.y_index)
    )
    if positive_x_element is not None:
        positive_x_wise_conduction: float = (
            pvt_collector.pv.conductivity  # [W/m*K]
            * pvt_collector.pv.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + positive_x_element.width))  # [m]
        )
    else:
        positive_x_wise_conduction = 0
    logger.debug(
        "Positive PV x-wise conduction term: %s W/K", positive_x_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index - 1, element.y_index)
    )
    if negative_x_element is not None:
        negative_x_wise_conduction: float = (
            pvt_collector.pv.conductivity  # [W/m*K]
            * pvt_collector.pv.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + negative_x_element.width))  # [m]
        )
    else:
        negative_x_wise_conduction = 0
    logger.debug(
        "Negative PV x-wise conduction term: %s W/K", negative_x_wise_conduction
    )

    # Compute the overall x-wise conduction term.
    x_wise_conduction = positive_x_wise_conduction + negative_x_wise_conduction
    logger.debug("PV x-wise conduction term: %s W/K", x_wise_conduction)

    # Compute the positive conductive term based on the next element along.
    positive_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index + 1)
    )
    if positive_y_element is not None:
        positive_y_wise_conduction: float = (
            pvt_collector.pv.conductivity  # [W/m*K]
            * pvt_collector.pv.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + positive_y_element.length))  # [m]
        )
    else:
        positive_y_wise_conduction = 0
    logger.debug(
        "Positive PV y-wise conduction term: %s W/K", positive_y_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index - 1)
    )
    if negative_y_element is not None:
        negative_y_wise_conduction: float = (
            pvt_collector.pv.conductivity  # [W/m*K]
            * pvt_collector.pv.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + negative_y_element.length))  # [m]
        )
    else:
        negative_y_wise_conduction = 0
    logger.debug(
        "Negative PV y-wise conduction term: %s W/K", negative_y_wise_conduction
    )

    # Compute the overall y-wise conduction term.
    y_wise_conduction = positive_y_wise_conduction + negative_y_wise_conduction
    logger.debug("PV y-wise conduction term: %s W/K", y_wise_conduction)

    # Compute the T_pv(i, j) term
    row_equation[
        index_handler.index_from_element_coordinates(
            number_of_x_elements,
            pvt_collector,
            TemperatureName.pv,
            element.x_index,
            element.y_index,
        )
    ] = (
        pv_internal_energy
        + x_wise_conduction
        + y_wise_conduction
        + pv_to_air_conduction
        + pv_to_sky_radiation
        + pv_to_glass_radiation
        + pv_to_glass_conduction
        + pv_to_absorber_conduction
    )

    # Compute the T_pv(i+1, j) term provided that that element exists.
    if element.x_index + 1 < number_of_x_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index + 1,
                element.y_index,
            )
        ] = (
            -1 * positive_x_wise_conduction
        )

    # Compute the T_pv(i-1, j) term provided that that element exists.
    if element.x_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index - 1,
                element.y_index,
            )
        ] = (
            -1 * negative_x_wise_conduction
        )

    # Compute the T_pv(i, j+1) term provided that that element exists.
    if element.y_index + 1 < number_of_y_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index,
                element.y_index + 1,
            )
        ] = (
            -1 * positive_y_wise_conduction
        )

    # Compute the T_pv(i, j-1) term provided that that element exists.
    if element.y_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index,
                element.y_index - 1,
            )
        ] = (
            -1 * negative_y_wise_conduction
        )

    # Compute the T_g(i, j) term provided that there is a glass layer present.
    if element.glass:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index,
                element.y_index,
            )
        ] = -1 * (pv_to_glass_conduction + pv_to_glass_radiation)

    # Compute the T_A(i, j) term provided that there is a absorber layer present.
    if element.absorber:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index,
                element.y_index,
            )
        ] = (
            -1 * pv_to_absorber_conduction
        )

    solar_thermal_resultant_vector_absorbtion_term = (
        pvt_collector.pv_transmissivity_absorptivity_product
        * weather_conditions.irradiance  # [W/m^2]
        * element.width  # [m]
        * element.length  # [m]
    ) - (
        weather_conditions.irradiance  # [W/m^2]
        * element.width  # [m]
        * element.length  # [m]
        * electrical_efficiency(
            pvt_collector,
            best_guess_temperature_vector[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.pv,
                    element.x_index,
                    element.y_index,
                )
            ],
        )
    )
    logger.debug(
        "PV solar thermal resultant vector term: %s W/K",
        solar_thermal_resultant_vector_absorbtion_term,
    )

    # Compute the resultant vector value.
    resultant_vector_value = (
        # Solar thermal absorption term.
        solar_thermal_resultant_vector_absorbtion_term  # [W]
        # Ambient temperature term.
        + pv_to_air_conduction * weather_conditions.ambient_temperature  # [W]
        # Sky temperature term.
        + pv_to_sky_radiation * weather_conditions.sky_temperature  # [W]
    )

    if operating_mode.dynamic:
        resultant_vector_value += (
            # Internal energy change
            pv_internal_energy  # [W/K]
            * previous_temperature_vector[  # type: ignore
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.pv,
                    element.x_index,
                    element.y_index,
                )
            ]
        )

    return row_equation, resultant_vector_value



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/adhesive_.py - Represents adhesive layer within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The adhesive_ module for the PV-T model.

This module represents a adhesive layer within a PV-T panel.

In addition to the upper glass cover represnted, there is scope for a small adhesive
layer placed in direct contact with the upper surface of the PV module.

"""

from .__utils__ import MicroLayer

__all__ = ("Adhesive",)


class Adhesive(MicroLayer):
    """
    Represents an Adhesive layer within the panel.

    """



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/pipe.py - Represents a pipe within the system.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The pipe module for the PV-T model.

This module represents a pipe within the PV-T system.

"""

from dataclasses import dataclass

__all__ = ("Pipe",)


@dataclass
class Pipe:
    """
    Represents a pipe within the PVT system.

    .. attribute:: temperature
        The temperature of the pipe, measured in Kelvin.

    """

    temperature: float



#!/usr/bin/python3.7
########################################################################################
# matrix.py - The matrix coefficient solver for the PV-T model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The matrix coefficient solver for the PV-T model.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

The equations in the system can be expressed as the matrix equation
    A * T = B
where
    A is the matrix computed and returned by this function;
    T is a vector containing the various temperatures;
    and B is the "resultant vector," computed elsewhere.

The equations represented by the rows of the matrix are:

The temperatures represented in the vector T are:

"""

from typing import List, Optional, Tuple, Union

import logging
import numpy

from tqdm import tqdm

from .. import exchanger, index_handler, tank
from ..pvt_collector import pvt

from ...__utils__ import (
    BColours,
    TemperatureName,
    OperatingMode,
    ProgrammerJudgementFault,
)
from ..__utils__ import WeatherConditions
from ..physics_utils import (
    convective_heat_transfer_coefficient_of_water,
    radiative_heat_transfer_coefficient,
)
from ..pvt_collector.element import Element
from ..pvt_collector.physics_utils import (
    glass_absorber_air_gap_resistance,
    glass_glass_air_gap_resistance,
    glass_pv_air_gap_resistance,
)

from .absorber import calculate_absorber_equation
from .continuity import (
    calculate_decoupled_system_continuity_equation,
    calculate_fluid_continuity_equation,
    calculate_system_continuity_equations,
)

from .glass import calculate_glass_equation
from .htf import calculate_htf_continuity_equation, calculate_htf_equation
from .pipe import calculate_pipe_equation
from .pv import calculate_pv_equation
from .tank import calculate_tank_continuity_equation, calculate_tank_equation
from .upper_glass import calculate_upper_glass_equation

__all__ = ("calculate_matrix_equation",)


####################
# Internal methods #
####################


def _boundary_condition_equations(
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    pvt_collector: pvt.PVT,
) -> List[Tuple[List[float], float]]:
    """
    Returns matrix rows and resultant vector values representing boundary conditions.

    These inluce:
        - there is no x temperature gradient at the "left" and "right" edges of the
          panel: physically, this means that the temperatures of cells near the x = 0
          and x = W boundaries are equal. These are represented as equations 1 (at the
          x=0 boundary) and 2 (at the x=W boundary);
        - there is no y temperature gradient along the "bottom" or "top" edges of the
          panel: physically, this means taht the temperatures of cells near the y = 0
          and y = H boundaries are equal. These are represented as equations 3 (at the
          y=0 boundary) and 4 (at the y=H boundary).

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `list` of `tuple`s containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    equations: List[Tuple[List[float], float]] = list()

    # Work through each layer, applying the boundary conditions.
    for temperature_name in {
        TemperatureName.glass,
        TemperatureName.pv,
        TemperatureName.absorber,
    }:
        # Work along both "left" and "right" edges, applying the boundary conditions.
        for y_coord in range(number_of_y_elements):
            # Equation 1: Zero temperature gradient at x=0.
            row_equation: List[float] = [0] * number_of_temperatures
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    0,
                    y_coord,
                )
            ] = -1
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    1,
                    y_coord,
                )
            ] = 1
            equations.append((row_equation, 0))
            # Equation 2: Zero temperature gradient at x=W.
            row_equation = [0] * number_of_temperatures
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    number_of_x_elements - 1,
                    y_coord,
                )
            ] = -1
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    number_of_x_elements - 2,
                    y_coord,
                )
            ] = 1
            equations.append((row_equation, 0))

        # Work along both "top" and "bottom" edges, applying the boundary conditions.
        for x_coord in range(number_of_x_elements):
            # Equation 3: Zero temperature gradient at y=0.
            row_equation = [0] * number_of_temperatures
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    x_coord,
                    0,
                )
            ] = -1
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    x_coord,
                    1,
                )
            ] = 1
            equations.append((row_equation, 0))
            # Equation 4: Zero temperature gradient at y=H.
            row_equation = [0] * number_of_temperatures
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    x_coord,
                    number_of_y_elements - 1,
                )
            ] = -1
            row_equation[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    x_coord,
                    number_of_y_elements - 2,
                )
            ] = 1
            equations.append((row_equation, 0))

    return equations


def _calculate_downward_glass_terms(
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    element: Element,
    logger: logging.Logger,
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    weather_conditions: WeatherConditions,
) -> Tuple[float, float]:
    """
    Calculates the downward glass terms.

    :return:
        A `tuple` containing:
        - the downward conduction from the glass layer to the layer below,
        - the downward radiation from the glass layer to the layer below.

    """

    # If the PV layer is below the glass layer.
    if element.glass and element.pv:
        if pvt_collector.glass is None:
            raise ProgrammerJudgementFault(
                "{}Element {} has a glass layer but no glass data supplied ".format(
                    BColours.FAIL, element
                )
                + "in the PV-T data file.{}".format(BColours.ENDC)
            )

        glass_downward_conduction = (
            element.width
            * element.length
            / glass_pv_air_gap_resistance(
                pvt_collector,
                0.5
                * (
                    best_guess_temperature_vector[
                        index_handler.index_from_element_coordinates(
                            number_of_x_elements,
                            pvt_collector,
                            TemperatureName.pv,
                            element.x_index,
                            element.y_index,
                        )
                    ]
                    + best_guess_temperature_vector[
                        index_handler.index_from_element_coordinates(
                            number_of_x_elements,
                            pvt_collector,
                            TemperatureName.glass,
                            element.x_index,
                            element.y_index,
                        )
                    ]
                ),
                weather_conditions,
            )
        )
        logger.debug("Glass to pv conduction %s W/K", glass_downward_conduction)

        glass_downward_radiation = (
            element.width
            * element.length
            * radiative_heat_transfer_coefficient(
                destination_emissivity=pvt_collector.pv.emissivity,
                destination_temperature=best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.pv,
                        element.x_index,
                        element.y_index,
                    )
                ],
                source_emissivity=pvt_collector.glass.emissivity,
                source_temperature=best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.glass,
                        element.x_index,
                        element.y_index,
                    )
                ],
            )
        )
        logger.debug("Glass to pv radiation %s W/K", glass_downward_radiation)

        return glass_downward_conduction, glass_downward_radiation

    # If the absorber layer is below the glass layer.
    if element.glass and element.absorber:
        if pvt_collector.glass is None:
            raise ProgrammerJudgementFault(
                "{}Element {} has a glass layer but no glass data supplied ".format(
                    BColours.FAIL, element
                )
                + "in the PV-T data file.{}".format(BColours.ENDC)
            )

        glass_downward_conduction = (
            element.width
            * element.length
            / glass_absorber_air_gap_resistance(
                pvt_collector,
                0.5
                * (
                    best_guess_temperature_vector[
                        index_handler.index_from_element_coordinates(
                            number_of_x_elements,
                            pvt_collector,
                            TemperatureName.absorber,
                            element.x_index,
                            element.y_index,
                        )
                    ]
                    + best_guess_temperature_vector[
                        index_handler.index_from_element_coordinates(
                            number_of_x_elements,
                            pvt_collector,
                            TemperatureName.glass,
                            element.x_index,
                            element.y_index,
                        )
                    ]
                ),
                weather_conditions,
            )
        )
        logger.debug("Glass to absorber conduction %s W/K", glass_downward_conduction)

        glass_downward_radiation = (
            element.width
            * element.length
            * radiative_heat_transfer_coefficient(
                destination_emissivity=pvt_collector.absorber.emissivity,
                destination_temperature=best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.absorber,
                        element.x_index,
                        element.y_index,
                    )
                ],
                source_emissivity=pvt_collector.glass.emissivity,
                source_temperature=best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.glass,
                        element.x_index,
                        element.y_index,
                    )
                ],
            )
        )
        logger.debug("Glass to absorber radiation %s W/K", glass_downward_radiation)

        return glass_downward_conduction, glass_downward_radiation

    # Otherwise, there is no glass layer, so these terms are zero.
    logger.debug(
        "No glass-to-pv conduction or radiation because {} ".format(
            " and ".join(
                {
                    entry
                    for entry in {
                        "glass" if not element.glass else None,
                        "pv" if not element.pv else None,
                    }
                    if entry is not None
                }
            )
        )
        + "layer(s) not present."
    )
    return 0, 0


def _calculate_downward_upper_glass_terms(
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    element: Element,
    logger: logging.Logger,
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    weather_conditions: WeatherConditions,
) -> Tuple[float, float]:
    """
    Calculates the downward conductive and radiative terms between the glass layers.

    :return:
        A `tuple` containing:
        - the conductive term between the two glass layers,
        - the radiative term between the two glass layers.

    """

    # If there is no double glazing, then these terms are zero.
    if not element.upper_glass:
        return 0, 0

    if not pvt_collector.upper_glass:
        raise ProgrammerJudgementFault(
            "{}Cannot compute double-glazing without `upper_glass` layer ".format(
                BColours.FAIL
            )
            + "present.{}".format(BColours.ENDC)
        )

    if not pvt_collector.glass:
        raise ProgrammerJudgementFault(
            "{}Cannot compute double-glazing without `glass` layer present.{}".format(
                BColours.FAIL, BColours.ENDC
            )
        )

    # Otherwise, calculate the terms.
    upper_glass_downward_conduction = (
        element.width
        * element.length
        / glass_glass_air_gap_resistance(
            pvt_collector,
            0.5
            * (
                best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.upper_glass,
                        element.x_index,
                        element.y_index,
                    )
                ]
                + best_guess_temperature_vector[
                    index_handler.index_from_element_coordinates(
                        number_of_x_elements,
                        pvt_collector,
                        TemperatureName.glass,
                        element.x_index,
                        element.y_index,
                    )
                ]
            ),
            weather_conditions,
        )
    )
    logger.debug(
        "Upper glass to glass conduction %s W/K", upper_glass_downward_conduction
    )

    upper_glass_downward_radiation = (
        element.width
        * element.length
        * radiative_heat_transfer_coefficient(
            destination_emissivity=pvt_collector.glass.emissivity,
            destination_temperature=best_guess_temperature_vector[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.glass,
                    element.x_index,
                    element.y_index,
                )
            ],
            source_emissivity=pvt_collector.upper_glass.emissivity,
            source_temperature=best_guess_temperature_vector[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.upper_glass,
                    element.x_index,
                    element.y_index,
                )
            ],
        )
    )
    logger.debug(
        "Upper glass to glass radiation %s W/K", upper_glass_downward_radiation
    )

    return upper_glass_downward_conduction, upper_glass_downward_radiation


##################
# Public methods #
##################


def calculate_matrix_equation(  # pylint: disable=too-many-branches
    *,
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    logger: logging.Logger,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    weather_conditions: WeatherConditions,
    collector_input_temperature: Optional[float] = None,
    heat_exchanger: Optional[exchanger.Exchanger] = None,
    hot_water_load: Optional[float] = None,
    hot_water_tank: Optional[tank.Tank] = None,
    previous_temperature_vector: Optional[numpy.ndarray] = None,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Calculates and returns both the matrix and resultant vector for the matrix equation.

    :return:
        A `tuple` containing both the matrix "A" and resultant vector "B" for the matrix
        equation representing the temperature equations.

    """

    logger.info("Matrix module called: calculating matrix and resultant vector.")
    logger.info(
        "A %s and %s matrix will be computed.",
        "dynamic" if operating_mode.dynamic else "steady-state",
        "decoupled" if operating_mode.decoupled else "coupled",
    )

    # Instantiate an empty matrix and array based on the number of temperatures present.
    matrix = numpy.zeros([0, number_of_temperatures])
    resultant_vector = numpy.zeros([0, 1])

    for element_coordinates, element in tqdm(
        pvt_collector.elements.items(),
        desc="computing matrix equation",
        unit="element",
        leave=False,
    ):
        logger.debug("Calculating equations for element %s", element_coordinates)
        # Compute the various shared values.
        (
            upper_glass_downward_conduction,
            upper_glass_downward_radiation,
        ) = _calculate_downward_upper_glass_terms(
            best_guess_temperature_vector,
            element,
            logger,
            number_of_x_elements,
            pvt_collector,
            weather_conditions,
        )

        (
            glass_downward_conduction,
            glass_downward_radiation,
        ) = _calculate_downward_glass_terms(
            best_guess_temperature_vector,
            element,
            logger,
            number_of_x_elements,
            pvt_collector,
            weather_conditions,
        )

        pv_to_absorber_conduction = (
            element.width
            * element.length
            / pvt_collector.pv_to_absorber_thermal_resistance
        )
        logger.debug("PV to absorber conduction: %s W/K", pv_to_absorber_conduction)

        absorber_to_pipe_conduction = (
            (
                element.width  # [m]
                * element.length  # [m]
                * pvt_collector.bond.conductivity  # [W/m*K]
                / pvt_collector.bond.thickness  # [m]
            )
            if element.pipe
            else 0
        )
        logger.debug("Absorber to pipe conduction: %s W/K", absorber_to_pipe_conduction)

        if element.upper_glass:
            (
                upper_glass_equation,
                upper_glass_resultant_value,
            ) = calculate_upper_glass_equation(
                best_guess_temperature_vector,
                logger,
                number_of_temperatures,
                number_of_x_elements,
                number_of_y_elements,
                operating_mode,
                previous_temperature_vector,
                pvt_collector,
                resolution,
                element,
                upper_glass_downward_conduction,
                upper_glass_downward_radiation,
                weather_conditions,
            )
            logger.debug(
                "Upper glass equation for element %s computed:\nEquation: %s\n"
                "Resultant value: %s W",
                element.coordinates,
                ", ".join([f"{value:.3f} W/K" for value in upper_glass_equation]),
                upper_glass_resultant_value,
            )
            matrix = numpy.vstack((matrix, upper_glass_equation))
            resultant_vector = numpy.vstack(
                (resultant_vector, upper_glass_resultant_value)
            )

        if element.glass:
            glass_equation, glass_resultant_value = calculate_glass_equation(
                best_guess_temperature_vector,
                glass_downward_conduction,
                glass_downward_radiation,
                logger,
                number_of_temperatures,
                number_of_x_elements,
                number_of_y_elements,
                operating_mode,
                previous_temperature_vector,
                pvt_collector,
                resolution,
                element,
                upper_glass_downward_conduction,
                upper_glass_downward_radiation,
                weather_conditions,
            )
            logger.debug(
                "Glass equation for element %s computed:\nEquation: %s\nResultant value: %s W",
                element.coordinates,
                ", ".join([f"{value:.3f} W/K" for value in glass_equation]),
                glass_resultant_value,
            )
            matrix = numpy.vstack((matrix, glass_equation))
            resultant_vector = numpy.vstack((resultant_vector, glass_resultant_value))

        if element.pv:
            pv_equation, pv_resultant_value = calculate_pv_equation(
                best_guess_temperature_vector,
                logger,
                number_of_temperatures,
                number_of_x_elements,
                number_of_y_elements,
                operating_mode,
                previous_temperature_vector,
                pv_to_absorber_conduction,
                glass_downward_conduction,
                glass_downward_radiation,
                pvt_collector,
                resolution,
                element,
                weather_conditions,
            )
            logger.debug(
                "PV equation for element %s computed:\nEquation: %s\nResultant value: %s W",
                element.coordinates,
                ", ".join([f"{value:.3f} W/K" for value in pv_equation]),
                pv_resultant_value,
            )
            matrix = numpy.vstack((matrix, pv_equation))
            resultant_vector = numpy.vstack((resultant_vector, pv_resultant_value))

        if element.absorber:
            absorber_equation, absorber_resultant_value = calculate_absorber_equation(
                absorber_to_pipe_conduction,
                best_guess_temperature_vector,
                glass_downward_conduction,
                glass_downward_radiation,
                logger,
                number_of_pipes,
                number_of_temperatures,
                number_of_x_elements,
                number_of_y_elements,
                operating_mode,
                previous_temperature_vector,
                pv_to_absorber_conduction,
                pvt_collector,
                resolution,
                element,
                weather_conditions,
            )
            logger.debug(
                "Collector equation for element %s computed:\nEquation: %s\nResultant value: %s W",
                element.coordinates,
                ", ".join([f"{value:.3f} W/K" for value in absorber_equation]),
                absorber_resultant_value,
            )
            matrix = numpy.vstack((matrix, absorber_equation))
            resultant_vector = numpy.vstack(
                (resultant_vector, absorber_resultant_value)
            )

        # Only calculate the pipe equations if the element has an associated pipe.
        if not element.pipe:
            logger.debug("3 equations for element %s", element_coordinates)
            continue

        pipe_to_htf_heat_transfer = (
            element.length  # [m]
            * numpy.pi
            * pvt_collector.absorber.inner_pipe_diameter  # [m]
            * convective_heat_transfer_coefficient_of_water(
                best_guess_temperature_vector[
                    index_handler.index_from_pipe_coordinates(
                        number_of_pipes,
                        number_of_x_elements,
                        element.pipe_index,  # type: ignore
                        pvt_collector,
                        TemperatureName.htf,
                        element.y_index,
                    )
                ],
                pvt_collector,
                weather_conditions,
            )  # [W/m^2*K]
        )
        logger.debug("Pipe to HTF heat transfer: %s W/K", pipe_to_htf_heat_transfer)

        pipe_equation, pipe_resultant_value = calculate_pipe_equation(
            absorber_to_pipe_conduction,
            best_guess_temperature_vector,
            logger,
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            operating_mode,
            pipe_to_htf_heat_transfer,
            previous_temperature_vector,
            pvt_collector,
            resolution,
            element,
            weather_conditions,
        )
        logger.debug(
            "Pipe equation for element %s computed:\nEquation: %s\nResultant value: %s W",
            element.coordinates,
            ", ".join([f"{value:.3f} W/K" for value in pipe_equation]),
            pipe_resultant_value,
        )
        matrix = numpy.vstack((matrix, pipe_equation))
        resultant_vector = numpy.vstack((resultant_vector, pipe_resultant_value))

        htf_equation, htf_resultant_value = calculate_htf_equation(
            best_guess_temperature_vector,
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            operating_mode,
            pipe_to_htf_heat_transfer,
            previous_temperature_vector,
            pvt_collector,
            resolution,
            element,
        )
        logger.debug(
            "HTF equation for element %s computed:\nEquation: %s\nResultant value: %s W",
            element.coordinates,
            ", ".join([f"{value:.3f} W/K" for value in htf_equation]),
            htf_resultant_value,
        )
        matrix = numpy.vstack((matrix, htf_equation))
        resultant_vector = numpy.vstack((resultant_vector, htf_resultant_value))

        htf_equation, htf_resultant_value = calculate_htf_continuity_equation(
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            pvt_collector,
            element,
        )
        logger.debug(
            "HTF definition equation for element %s computed:\nEquation: %s\nResultant value: %s W",
            element.coordinates,
            ", ".join([f"{value:.3f} W/K" for value in htf_equation]),
            htf_resultant_value,
        )
        matrix = numpy.vstack((matrix, htf_equation))
        resultant_vector = numpy.vstack((resultant_vector, htf_resultant_value))

        # Fluid continuity equations only need to be computed if there exist multiple
        # connected elements.
        if element.y_index >= number_of_y_elements - 1:
            logger.debug("6 equations for element %s", element_coordinates)
            continue

        (
            fluid_continuity_equation,
            fluid_continuity_resultant_value,
        ) = calculate_fluid_continuity_equation(
            number_of_pipes,
            number_of_temperatures,
            number_of_x_elements,
            pvt_collector,
            element,
        )
        logger.debug(
            "Fluid continuity equation for element %s computed:\nEquation: %s\n"
            "Resultant value: %s W",
            element.coordinates,
            ", ".join([f"{value:.3f} W/K" for value in fluid_continuity_equation]),
            fluid_continuity_resultant_value,
        )
        matrix = numpy.vstack((matrix, fluid_continuity_equation))
        resultant_vector = numpy.vstack(
            (resultant_vector, fluid_continuity_resultant_value)
        )
        logger.debug("7 equations for element %s", element_coordinates)

    # # Calculate the system boundary condition equations.
    # boundary_condition_equations = _boundary_condition_equations(
    #     number_of_temperatures, number_of_x_elements, number_of_y_elements
    # )

    # for equation, resultant_value in boundary_condition_equations:
    #     logger.debug(
    #         "Boundary condition equation computed:\nEquation: %s\nResultant value: %s W",
    #         ", ".join([f"{value:.3f} W/K" for value in equation]),
    #         resultant_value,
    #     )
    #     matrix = numpy.vstack((matrix, equation))
    #     resultant_vector = numpy.vstack((resultant_vector, resultant_value))
    #     # if len(matrix) == number_of_temperatures:
    #     #     return matrix, resultant_vector

    # If the system is decoupled, do not compute and add the tank-related equations.
    if operating_mode.decoupled:
        if collector_input_temperature is None:
            raise ProgrammerJudgementFault(
                "{}No collector input temperature passed to the matrix module ".format(
                    BColours.FAIL
                )
                + "for decoupled operation.{}".format(BColours.ENDC)
            )
        decoupled_system_continuity_equations = (
            calculate_decoupled_system_continuity_equation(
                collector_input_temperature,
                number_of_pipes,
                number_of_temperatures,
                number_of_x_elements,
                number_of_y_elements,
                pvt_collector,
            )
        )
        for equation, resultant_value in decoupled_system_continuity_equations:
            logger.debug(
                "System continuity equation computed:\nEquation: %s\nResultant value: %s W",
                ", ".join([f"{value:.3f} W/K" for value in equation]),
                resultant_value,
            )
            matrix = numpy.vstack((matrix, equation))
            resultant_vector = numpy.vstack((resultant_vector, resultant_value))
        logger.info("Matrix equation computed, matrix dimensions: %s", matrix.shape)

        return matrix, resultant_vector

    if (
        heat_exchanger is None
        or hot_water_load is None
        or hot_water_tank is None
        or previous_temperature_vector is None
    ):
        raise ProgrammerJudgementFault(
            "{}Insufficient parameters for dynamic run:{}{}{}{}{}".format(
                " Heat exchanger missing." if heat_exchanger is None else "",
                " Hot water load missing." if hot_water_load is None else "",
                " Hot-water tank missing." if hot_water_tank is None else "",
                " Previous temperature vector is missing."
                if previous_temperature_vector is None
                else "",
                BColours.FAIL,
                BColours.ENDC,
            )
        )

    # Calculate the tank equations.
    equation, resultant_value = calculate_tank_equation(
        best_guess_temperature_vector,
        heat_exchanger,
        hot_water_load,
        hot_water_tank,
        logger,
        number_of_temperatures,
        previous_temperature_vector,
        pvt_collector,
        resolution,
        weather_conditions,
    )
    logger.debug(
        "Tank equation computed:\nEquation: %s\nResultant value: %s W",
        equation,
        resultant_value,
    )
    matrix = numpy.vstack((matrix, equation))
    resultant_vector = numpy.vstack((resultant_vector, resultant_value))

    # Calculate the tank continuity equation.
    equation, resultant_value = calculate_tank_continuity_equation(
        best_guess_temperature_vector,
        heat_exchanger,
        number_of_temperatures,
        pvt_collector,
    )
    logger.debug(
        "Tank continuity equation computed:\nEquation: %s\nResultant value: %s W",
        equation,
        resultant_value,
    )
    matrix = numpy.vstack((matrix, equation))
    resultant_vector = numpy.vstack((resultant_vector, resultant_value))
    logger.debug("2 tank equations computed.")

    # Compute the system continuity equations and assign en masse.
    system_continuity_equations = calculate_system_continuity_equations(
        number_of_pipes,
        number_of_temperatures,
        number_of_x_elements,
        number_of_y_elements,
        previous_temperature_vector,
        pvt_collector,
    )
    logger.debug(
        "%s system continuity equations computed.", len(system_continuity_equations)
    )

    for equation, resultant_value in system_continuity_equations:
        logger.debug(
            "System continuity equation computed:\nEquation: %s\nResultant value: %s W",
            ", ".join([f"{value:.3f} W/K" for value in equation]),
            resultant_value,
        )
        matrix = numpy.vstack((matrix, equation))
        resultant_vector = numpy.vstack((resultant_vector, resultant_value))

    logger.info("Matrix equation computed, matrix dimensions: %s", matrix.shape)

    return matrix, resultant_vector
